import pyomo.gdp as gdp
import pyomo.environ as pyo
import numpy as np
import pandas as pd

from ..component import ModelComponent
from ..utilities import (
    annualize,
    set_discount_rate,
    link_full_resolution_to_clustered,
    determine_variable_scaling,
    determine_constraint_scaling,
)
from .utilities import set_capex_model
from .ccs import fit_ccs_coeff

import logging

log = logging.getLogger(__name__)


class Technology(ModelComponent):
    """
    Class to read and manage data for technologies

    This class is parent class to all generic and specific technologies. It creates
    the variables, parameters, constraints and sets of a technology.

    This function is extented in the generic/specific technology classes. It adds Sets,
    Parameters, Variables and Constraints that are common for all technologies. The
    following description is true for new technologies. For existing technologies a
    few adaptions are made (see below).
    When CCS is available, we add heat and electricity to the input carriers Set and
    CO2captured to the output carriers Set. Moreover, we create extra Parameters and
    Variables equivalent to the ones created for the technology, but specific for
    CCS. In addition, we create Variables that are the sum of the input, output,
    CAPEX and OPEX of the technology and of CCS. We calculate the emissions of the
    technology discounting already
    what is being captured by the CCS.

    **Set declarations:**

    - set_input_carriers: Set of input carriers
    - set_output_carriers: Set of output carriers

    If ccs is possible:

    - set_input_carriers_ccs: Set of CCS input carriers
    - set_output_carriers_ccs: Set of CCS output carriers

    ** Set declarations for time aggregation:**

    Three sets are declared for each technology. These are required for time
    averaging algorithms:

    - set_t_full: set of all time steps before clustering
    - set_t_performance: set of time steps, on which the technology performance is
      based on
    - set_t_global: set of time steps, on which the energy balance is based on

    **Parameter declarations:**

    The following is a list of declared pyomo parameters.

    - para_size_min: minimal possible size
    - para_size_max: maximal possible size
    - para_unit_capex: investment costs per unit
    - para_unit_capex_annual: Unit CAPEX annualized (annualized from given data on
      up-front CAPEX, lifetime and discount rate)
    - para_fix_capex: fixed costs independent of size
    - para_fix_capex_annual: Fixed CAPEX annualized (annualized from given data on
      up-front CAPEX, lifetime and discount rate)
    - para_opex_variable: operational cost EUR/output or input
    - para_opex_fixed: fixed opex as fraction of annualized capex
    - para_tec_emissionfactor: emission factor per output or input

    If ccs is possible:

    - para_size_min_ccs: minimal possible size
    - para_size_max_ccs: maximal possible size
    - para_unit_capex_annual_ccs: investment costs per unit (annualized from given data on up-front CAPEX, lifetime
      and discount rate)
    - para_fix_capex_annual_ccs: Fixed CAPEX annualized (annualized from given data on
      up-front CAPEX, lifetime and discount rate)
    - para_opex_variable_ccs: operational cost EUR/output or input
    - para_opex_fixed_ccs: fixed opex as fraction of annualized capex

    For existing technologies:

    - para_size_initial: initial size
    - para_decommissioning_cost: Decommissioning cost

    **Variable declarations:**

    - var_size: Size of the technology, can be integer or continuous
    - var_input: input to the technology, defined for each input carrier and time slice
    - var_output: output of the technology, defined for each output carrier and time
      slice
    - var_input_tot: input aggregation of technology and CCS input
    - var_output_tot: output aggregation of technology and CCS output
    - var_capex: annualized investment of the technology
    - var_opex_variable: variable operation costs, defined for each time slice
    - var_opex_fixed: fixed operational costs
    - var_capex_tot: aggregation of technology and CCS capex
    - var_capex_aux: auxiliary variable to calculate the fixed opex of existing technologies
    - var_opex_variable_tot: aggregation of technology and CCS opex variable, defined for
      each time slice
    - var_opex_fixed_tot: aggregation of technology and CCS opex fixed
    - var_tec_emissions_pos: positive emissions, defined per time slice
    - var_tec_emissions_neg: negative emissions, defined per time slice

    If ccs is possible:

    - var_size_ccs: Size of CCS
    - var_input_ccs: input to the CCS component, defined for each CCS input carrier and
      time slice
    - var_output_ccs: output from the CCS component, defined for each CCS output carrier
      and time slice
    - var_capex_ccs: annualized investment of CCS
    - var_capex_aux_ccs: auxiliary variable to calculate the fixed opex of existing CCS
    - var_opex_variable_ccs: variable operation costs, defined for each time slice
    - var_opex_fixed_ccs: fixed operational costs

    **Constraint declarations**

    - For new technologies, CAPEX, can be linear (for ``capex_model == 1``), piecewise
      linear (for ``capex_model == 2``) or linear with a fixed cost when the
      technology is installed (for ``capex_model == 3``). The capex model can
      also be overwritten in the children technology classes  (for ``capex_model ==
      4``). Linear is defined as:

        .. math::
            capex_{aux} = size * capex_{unitannual}

      while linear with fixed installation costs is defined as. Note that capex_aux is
      zero if the technology is not installed:

        .. math::
            capex_{aux} = size * capex_{unitannual} + capex_{fixed}

      Existing technologies, i.e. existing = 1, can be decommissioned (decommission = 1) or not (decommission = 0).
      For technologies that cannot be decommissioned, the size is fixed to the size
      given in the technology data.
      For technologies that can be decommissioned, the size can be smaller or equal to the initial size. Reducing the
      size comes at the decommissioning costs specified in the economics of the technology.
      The fixed opex is calculated by determining the capex that the technology would have costed if newly build and
      then taking the respective opex_fixed share of this. This is done with the auxiliary variable capex_aux.

    - For existing technologies that can be decommissioned, the CAPEX equal to
      the decommissioning costs:

        .. math::
            capex = (size_{initial} - size) * decommissioningcost

    - Variable OPEX: variable opex is defined in terms of the input, with the
      exception of DAC_Adsorption, RES and CONV4, where it is defined per unit of
      output:

        .. math::
            opexvar_{t} = Input_{t, maincarrier} * opex_{var}

    - Fixed OPEX: defined as a fraction of annual CAPEX:

        .. math::
            opexfix = capex * opex_{fix}

    - Input aggregation: aggregates total input from technology
      and CCS. In case there is no CCS, input_ccs is zero:

        .. math::
            input_{t, car} + input_{CCS, t, car} = input_{tot, t, car}

    - Output aggregation: aggregates total output from technology
      and CCS. In case there is no CCS, output_ccs is zero:

        .. math::
            output_{t, car} + output_{CCS, t, car} = output_{tot, t, car}

    - Capex aggregation: aggregates capex of technology
      and CCS. In case there is no CCS, capex_ccs is zero:

        .. math::
            capex + capex_{CCS} = capex_{tot}

    - Opex variable aggregation: aggregates opex variable of technology
      and CCS. In case there is no CCS, var_opex_variable_ccs is zero:

        .. math::
            opex_{variable, t} + opex_{variable,CCS, t} =  opex_{variable,tot, t}

    - Opex fixed aggregation: aggregates opex fixed of technology
      and CCS. In case there is no CCS, opex_fixed_ccs is zero:

        .. math::
            opex_{fixed} + opex_{fixed,CCS} =  opex_{fixed,tot}

    - Emissions: depending if they are based on input or output and depending if
      emission factor is negative or positive

        .. math::
            emissions_{pos/neg,t} = output_{maincarrier, t} * emissionfactor_{pos}

    If CCS is possible the following constraints apply:

    - Input carriers are given by:

    .. math::
        input_{CCS, car} <= inputRatio_{carrier} * output_{CCS}/captureRate
    .. math::
        input_{tot, car} = inputTec_{car} + input_{CCS, car}

    - CO2 captured output is constrained by:

    .. math::
        output_{CCS} <= input(output)_{tec} * emissionFactor * captureRate

    - The total output are given by:

    .. math::
        output_{tot, car} = outputTec_{car} + output_{CCS, car}

    - Emissions of the technology are:

    .. math::
        emissions_{tec} = input(output)_{tec} * emissionFactor - output_{CCS}

    - CAPEX is given by

    .. math::
        CAPEX_{CCS} = Size_{CCS} * UnitCost_{CCS} + FixCost_{CCS}
    .. math::
        CAPEX_{tot} = CAPEX_{CCS} + CAPEX_{tec}

    - Fixed OPEX: defined as a fraction of annual CAPEX:

    .. math::
        OPEXfix_{CCS} = CAPEX_{CCS} * opex_{CCS}
    .. math::
        OPEX_{tot} = OPEX_{CCS} + OPEX_{tec}
    """

    def __init__(self, tec_data: dict):
        """
        Initializes technology class from technology data

        The technology name needs to correspond to the name of a JSON file in ./data/technology_data.

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        # Technology data
        self.ccs_data = None
        self.ccs_component = None

        # Modelling attributes
        self.input = None
        self.output = None
        self.set_t_full = None
        self.set_t_performance = None
        self.set_t_global = None
        self.sequence = None

        # Scaling factors
        self.scaling_factors = None
        if "ScalingFactors" in tec_data:
            self.scaling_factors = tec_data["ScalingFactors"]

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits technology performance (bounds and coefficients).

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        input_parameters = self.input_parameters
        time_independent = {}

        # Size
        time_independent["size_min"] = input_parameters.size_min
        if not self.existing:
            time_independent["size_max"] = input_parameters.size_max
        else:
            time_independent["size_max"] = input_parameters.size_initial
            time_independent["size_initial"] = input_parameters.size_initial

        # Emissions
        time_independent["emission_factor"] = input_parameters.performance_data[
            "emission_factor"
        ]

        # Other
        time_independent["rated_power"] = input_parameters.rated_power
        time_independent["min_part_load"] = input_parameters.min_part_load
        time_independent["standby_power"] = input_parameters.standby_power

        # Dynamics
        dynamics = {}
        dynamics_parameter = [
            "ramping_time",
            "ref_size",
            "ramping_const_int",
            "standby_power",
            "min_uptime",
            "min_downtime",
            "SU_time",
            "SD_time",
            "SU_load",
            "SD_load",
            "max_startups",
        ]
        for p in dynamics_parameter:
            if p in input_parameters.performance_data:
                dynamics[p] = input_parameters.performance_data[p]

        # Write to self
        self.processed_coeff.time_independent = time_independent
        self.processed_coeff.dynamics = dynamics

        # CCS
        if self.component_options.ccs_possible:
            co2_concentration = self.input_parameters.performance_data["ccs"][
                "co2_concentration"
            ]
            self.ccs_data["name"] = "CCS"
            self.ccs_data["tec_type"] = self.component_options.ccs_type
            self.ccs_component = fit_ccs_coeff(
                co2_concentration, self.ccs_data, climate_data
            )

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used. Overwritten in child classes
        """
        pass

    def _calculate_ccs_bounds(self):
        """
        Calculates bounds of CCS
        """
        time_steps = len(self.set_t_performance)

        # Calculate input and output bounds
        for car in self.ccs_component.component_options.input_carrier:
            self.ccs_component.bounds["input"][car] = np.column_stack(
                (
                    np.zeros(shape=(time_steps)),
                    np.ones(shape=(time_steps))
                    * self.ccs_component.processed_coeff.time_independent[
                        "input_ratios"
                    ][car],
                )
            )
        for car in self.ccs_component.component_options.output_carrier:
            self.ccs_component.bounds["output"][car] = np.column_stack(
                (
                    np.zeros(shape=(time_steps)),
                    np.ones(shape=(time_steps))
                    * self.ccs_component.processed_coeff.time_independent[
                        "capture_rate"
                    ],
                )
            )

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Construct the technology model with all required parameters, variable, sets,...

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        # LOG
        log_msg = f"\t - Adding Technology {self.name}"
        log.info(log_msg)

        # TECHNOLOGY DATA
        config = data["config"]

        # SET T
        self.set_t_full = set_t_full

        # MODELING TYPICAL DAYS
        technologies_modelled_with_full_res = config["optimization"]["typicaldays"][
            "technologies_with_full_res"
        ]["value"]

        if config["optimization"]["typicaldays"]["N"]["value"] == 0:
            # everything with full resolution
            self.component_options.modelled_with_full_res = True
            self.set_t_performance = set_t_full
            self.set_t_global = set_t_full
            self.sequence = list(self.set_t_performance)

        elif config["optimization"]["typicaldays"]["method"]["value"] == 1:
            # everything with reduced resolution
            self.component_options.modelled_with_full_res = False
            self.set_t_performance = set_t_clustered
            self.set_t_global = set_t_clustered
            self.sequence = list(self.set_t_performance)

        elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
            # resolution of balances is full, so interactions with them also need to
            # be full resolution
            self.set_t_global = set_t_full

            if (
                self.component_options.technology_model
                in technologies_modelled_with_full_res
            ):
                # technologies modelled with full resolution
                self.component_options.modelled_with_full_res = True
                self.component_options.lower_res_than_full = False
                self.set_t_performance = self.set_t_full
                self.sequence = list(self.set_t_performance)
            else:
                # technologies modelled with reduced resolution
                self.component_options.modelled_with_full_res = False
                self.component_options.lower_res_than_full = True
                self.set_t_performance = set_t_clustered
                self.sequence = data["k_means_specs"]["sequence"]

        # Coefficients
        if self.component_options.modelled_with_full_res:
            if config["optimization"]["timestaging"]["value"] == 0:
                self.processed_coeff.time_dependent_used = (
                    self.processed_coeff.time_dependent_full
                )
            else:
                self.processed_coeff.time_dependent_used = (
                    self.processed_coeff.time_dependent_averaged
                )
        else:
            self.processed_coeff.time_dependent_used = (
                self.processed_coeff.time_dependent_clustered
            )

        # CALCULATE BOUNDS
        self._calculate_bounds()

        # GENERAL TECHNOLOGY CONSTRAINTS
        b_tec = self._define_input_carriers(b_tec)
        b_tec = self._define_output_carriers(b_tec)
        b_tec = self._define_size(b_tec)
        b_tec = self._define_capex_parameters(b_tec, data)
        b_tec = self._define_capex_variables(b_tec, data)
        b_tec = self._define_capex_constraints(b_tec, data)
        b_tec = self._define_input(b_tec, data)
        b_tec = self._define_output(b_tec, data)

        # CLUSTERED DATA
        if (config["optimization"]["typicaldays"]["N"]["value"] == 0) or (
            config["optimization"]["typicaldays"]["method"]["value"] == 1
        ):
            # input/output to calculate performance is the same as var_input
            if b_tec.find_component("var_input"):
                self.input = b_tec.var_input
            if b_tec.find_component("var_output"):
                self.output = b_tec.var_output
        elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
            if (
                self.component_options.technology_model
                in technologies_modelled_with_full_res
            ):
                # input/output to calculate performance is the same as var_input
                if b_tec.find_component("var_input"):
                    self.input = b_tec.var_input
                if b_tec.find_component("var_output"):
                    self.output = b_tec.var_output
            else:
                # input/output to calculate performance has lower resolution
                b_tec = self._define_auxiliary_vars(b_tec, data)
                if b_tec.find_component("var_input"):
                    self.input = b_tec.var_input_aux
                if b_tec.find_component("var_output"):
                    self.output = b_tec.var_output_aux

        b_tec = self._define_opex(b_tec)

        # CCS and Emissions
        if self.component_options.ccs_possible:
            log_msg = f"\t - Adding CCS to Technology {self.name}"
            log.info(log_msg)
            self._calculate_ccs_bounds()
            if self.component_options.modelled_with_full_res:
                self.ccs_component.processed_coeff.time_dependent_used = (
                    self.ccs_component.processed_coeff.time_dependent_full
                )
            else:
                self.ccs_component.processed_coeff.time_dependent_used = (
                    self.ccs_component.processed_coeff.time_dependent_clustered
                )
            b_tec = self._define_ccs_performance(b_tec, data)
            b_tec = self._define_ccs_emissions(b_tec)
            b_tec = self._define_ccs_costs(b_tec, data)
            log_msg = f"\t - Adding CCS to Technology {self.name} completed"
            log.info(log_msg)

        else:
            b_tec = self._define_emissions(b_tec)

        # DYNAMICS
        if config["performance"]["dynamics"]["value"]:
            technologies_modelled_with_dynamics = ["CONV1", "CONV2", "CONV3"]
            if (
                self.component_options.technology_model
                in technologies_modelled_with_dynamics
            ):
                b_tec = self._define_dynamics(b_tec, data)
            else:
                log_msg = "Modeling dynamic constraints not enabled for technology type"
                log.warning(log_msg)

        else:
            if self.component_options.performance_function_type == 4:
                self.component_options.performance_function_type = 3
                log_msg = (
                    "Switching dynamics off for performance function type 4, "
                    "type changed to 3 for "
                ) + self.name

                log.warning(log_msg)

        # AGGREGATE ALL VARIABLES
        self._aggregate_input(b_tec)
        self._aggregate_output(b_tec)
        self._aggregate_cost(b_tec)

        return b_tec

    def _define_input_carriers(self, b_tec):
        """
        Defines the input carriers

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        b_tec.set_input_carriers = pyo.Set(
            initialize=self.component_options.input_carrier
        )

        if self.component_options.ccs_possible:
            b_tec.set_input_carriers_ccs = pyo.Set(
                initialize=self.ccs_component.component_options.input_carrier
            )
        else:
            b_tec.set_input_carriers_ccs = pyo.Set(initialize=[])

        b_tec.set_input_carriers_all = (
            b_tec.set_input_carriers_ccs | b_tec.set_input_carriers
        )

        return b_tec

    def _define_output_carriers(self, b_tec):
        """
        Defines the output carriers

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        b_tec.set_output_carriers = pyo.Set(
            initialize=self.component_options.output_carrier
        )

        if self.component_options.ccs_possible:
            b_tec.set_output_carriers_ccs = pyo.Set(
                initialize=self.ccs_component.component_options.output_carrier
            )
        else:
            b_tec.set_output_carriers_ccs = pyo.Set(initialize=[])

        b_tec.set_output_carriers_all = (
            b_tec.set_output_carriers_ccs | b_tec.set_output_carriers
        )

        return b_tec

    def _define_size(self, b_tec):
        """
        Defines variables and parameters related to technology size.

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        coeff_ti = self.processed_coeff.time_independent

        if self.component_options.size_is_int:
            size_domain = pyo.NonNegativeIntegers
        else:
            size_domain = pyo.NonNegativeReals

        b_tec.para_size_min = pyo.Param(
            domain=pyo.NonNegativeReals, initialize=coeff_ti["size_min"], mutable=True
        )
        b_tec.para_size_max = pyo.Param(
            domain=pyo.NonNegativeReals, initialize=coeff_ti["size_max"], mutable=True
        )

        if self.existing:
            b_tec.para_size_initial = pyo.Param(
                within=size_domain, initialize=coeff_ti["size_initial"]
            )

        if self.existing and not self.component_options.decommission:
            # Decommissioning is not possible, size fixed
            b_tec.var_size = pyo.Param(
                within=size_domain, initialize=b_tec.para_size_initial
            )
        else:
            # Decommissioning is possible, size variable
            b_tec.var_size = pyo.Var(
                within=size_domain, bounds=(b_tec.para_size_min, b_tec.para_size_max)
            )

        return b_tec

    def _define_capex_variables(self, b_tec, data: dict):
        """
        Defines variables related to technology capex.

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        config = data["config"]

        economics = self.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        capex_model = set_capex_model(config, economics)

        def calculate_max_capex():
            if capex_model == 1:
                max_capex = (
                    b_tec.para_size_max
                    * economics.capex_data["unit_capex"]
                    * annualization_factor
                )
                bounds = (0, max_capex)
            elif capex_model == 2:
                max_capex = (
                    b_tec.para_size_max
                    * max(economics.capex_data["piecewise_capex"]["bp_y"])
                    * annualization_factor
                )
                bounds = (0, max_capex)
            elif capex_model == 3:
                max_capex = (
                    b_tec.para_size_max * economics.capex_data["unit_capex"]
                    + economics.capex_data["fix_capex"]
                ) * annualization_factor
                bounds = (0, max_capex)
            else:
                bounds = None
            return bounds

        # CAPEX auxilliary (used to calculate theoretical CAPEX)
        # For new technologies, this is equal to actual CAPEX
        # For existing technologies it is used to calculate fixed OPEX
        b_tec.var_capex_aux = pyo.Var(bounds=calculate_max_capex())

        b_tec.var_capex = pyo.Var()

        return b_tec

    def _define_capex_parameters(self, b_tec, data):
        """
        Defines the capex parameters

        For capex model 1:
        - para_unit_capex
        - para_unit_capex_annual

        For capex model 2: defined with constraints
        For capex model 3:
        - para_unit_capex
        - para_fix_capex
        - para_unit_capex_annual
        - para_fix_capex_annual

        :param b_tec:
        :param data:
        :return:
        """
        config = data["config"]
        economics = self.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        capex_model = set_capex_model(config, economics)

        if capex_model == 1:
            b_tec.para_unit_capex = pyo.Param(
                domain=pyo.Reals,
                initialize=economics.capex_data["unit_capex"],
                mutable=True,
            )
            b_tec.para_unit_capex_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=annualization_factor * economics.capex_data["unit_capex"],
                mutable=True,
            )

        elif capex_model == 2:
            # This is defined in the constraints
            pass
        elif capex_model == 3:
            b_tec.para_unit_capex = pyo.Param(
                domain=pyo.Reals,
                initialize=economics.capex_data["unit_capex"],
                mutable=True,
            )
            b_tec.para_fix_capex = pyo.Param(
                domain=pyo.Reals,
                initialize=economics.capex_data["fix_capex"],
                mutable=True,
            )
            b_tec.para_unit_capex_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=annualization_factor * economics.capex_data["unit_capex"],
                mutable=True,
            )
            b_tec.para_fix_capex_annual = pyo.Param(
                domain=pyo.Reals,
                initialize=annualization_factor * economics.capex_data["fix_capex"],
                mutable=True,
            )
        else:
            # Defined in the technology subclass
            pass

        if self.existing and self.component_options.decommission:
            b_tec.para_decommissioning_cost = pyo.Param(
                domain=pyo.Reals, initialize=economics.decommission_cost, mutable=True
            )

        return b_tec

    def _define_capex_constraints(self, b_tec, data):
        """
        Defines constraints related to capex.
        """
        config = data["config"]
        economics = self.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        capex_model = set_capex_model(config, economics)

        if capex_model == 1:
            b_tec.const_capex_aux = pyo.Constraint(
                expr=b_tec.var_size * b_tec.para_unit_capex_annual
                == b_tec.var_capex_aux
            )
        elif capex_model == 2:
            self.big_m_transformation_required = 1
            bp_x = economics.capex_data["piecewise_capex"]["bp_x"]
            bp_y_annual = [
                y * annualization_factor
                for y in economics.capex_data["piecewise_capex"]["bp_y"]
            ]
            b_tec.const_capex_aux = pyo.Piecewise(
                b_tec.var_capex_aux,
                b_tec.var_size,
                pw_pts=bp_x,
                pw_constr_type="EQ",
                f_rule=bp_y_annual,
                pw_repn="SOS2",
            )
        elif capex_model == 3:
            self.big_m_transformation_required = 1
            s_indicators = range(0, 2)

            def init_installation(dis, ind):
                if ind == 0:  # tech not installed
                    dis.const_capex_aux = pyo.Constraint(expr=b_tec.var_capex_aux == 0)
                    dis.const_not_installed = pyo.Constraint(expr=b_tec.var_size == 0)
                else:  # tech installed
                    dis.const_capex_aux = pyo.Constraint(
                        expr=b_tec.var_size * b_tec.para_unit_capex_annual
                        + b_tec.para_fix_capex_annual
                        == b_tec.var_capex_aux
                    )

            b_tec.dis_installation = gdp.Disjunct(s_indicators, rule=init_installation)

            def bind_disjunctions(dis):
                return [b_tec.dis_installation[i] for i in s_indicators]

            b_tec.disjunction_installation = gdp.Disjunction(rule=bind_disjunctions)

        else:
            # Defined in the technology subclass
            pass

        # CAPEX
        if self.existing:
            if self.component_options.decommission:
                # technology can be decommissioned
                b_tec.para_decommissioning_cost = pyo.Param(
                    domain=pyo.Reals,
                    initialize=economics.decommission_cost,
                    mutable=True,
                )
                b_tec.const_capex = pyo.Constraint(
                    expr=b_tec.var_capex
                    == (b_tec.para_size_initial - b_tec.var_size)
                    * b_tec.para_decommissioning_cost
                )
            else:
                # technology cannot be decommissioned
                b_tec.const_capex = pyo.Constraint(expr=b_tec.var_capex == 0)
        else:
            b_tec.const_capex = pyo.Constraint(
                expr=b_tec.var_capex == b_tec.var_capex_aux
            )

        return b_tec

    def _define_input(self, b_tec, data: dict):
        """
        Defines input to a technology

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        # Technology related data
        c = self.processed_coeff.time_independent

        def init_input_bounds(bounds, t, car):
            return tuple(
                self.bounds["input"][car][self.sequence[t - 1] - 1, :]
                * c["size_max"]
                * c["rated_power"]
            )

        b_tec.var_input = pyo.Var(
            self.set_t_global,
            b_tec.set_input_carriers,
            within=pyo.NonNegativeReals,
            bounds=init_input_bounds,
        )

        return b_tec

    def _define_output(self, b_tec, data: dict):
        """
        Defines output to a technology

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        # Technology related data
        c = self.processed_coeff.time_independent

        def init_output_bounds(bounds, t, car):
            return tuple(
                self.bounds["output"][car][self.sequence[t - 1] - 1, :]
                * c["size_max"]
                * c["rated_power"]
            )

        b_tec.var_output = pyo.Var(
            self.set_t_global,
            b_tec.set_output_carriers,
            within=pyo.NonNegativeReals,
            bounds=init_output_bounds,
        )
        return b_tec

    def _define_opex(self, b_tec):
        """
        Defines variable and fixed OPEX

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        economics = self.economics

        # VARIABLE OPEX
        b_tec.para_opex_variable = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_variable, mutable=True
        )
        b_tec.var_opex_variable = pyo.Var(self.set_t_global)

        def init_opex_variable(const, t):
            """opexvar_{t} = Input_{t, maincarrier} * opex_{var}"""
            if (
                (self.component_options.technology_model == "RES")
                or (self.component_options.technology_model == "CONV4")
                or (self.component_options.technology_model == "DAC_Adsorption")
            ):
                opex_variable_based_on = b_tec.var_output[
                    t, b_tec.set_output_carriers[1]
                ]
            else:
                opex_variable_based_on = b_tec.var_input[
                    t, self.component_options.main_input_carrier
                ]
            return (
                opex_variable_based_on * b_tec.para_opex_variable
                == b_tec.var_opex_variable[t]
            )

        b_tec.const_opex_variable = pyo.Constraint(
            self.set_t_global, rule=init_opex_variable
        )

        # FIXED OPEX
        b_tec.para_opex_fixed = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_fixed, mutable=True
        )
        b_tec.var_opex_fixed = pyo.Var()
        b_tec.const_opex_fixed = pyo.Constraint(
            expr=b_tec.var_capex_aux * b_tec.para_opex_fixed == b_tec.var_opex_fixed
        )
        return b_tec

    def _define_emissions(self, b_tec):
        """
        Defines Emissions

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        c = self.processed_coeff.time_independent
        technology_model = self.component_options.technology_model
        emissions_based_on = self.component_options.emissions_based_on

        b_tec.para_tec_emissionfactor = pyo.Param(
            domain=pyo.Reals, initialize=c["emission_factor"]
        )
        b_tec.var_tec_emissions_pos = pyo.Var(
            self.set_t_global, within=pyo.NonNegativeReals
        )
        b_tec.var_tec_emissions_neg = pyo.Var(
            self.set_t_global, within=pyo.NonNegativeReals
        )

        if technology_model == "RES":
            # Set emissions to zero
            def init_tec_emissions_pos(const, t):
                return b_tec.var_tec_emissions_pos[t] == 0

            b_tec.const_tec_emissions_pos = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_pos
            )

            def init_tec_emissions_neg(const, t):
                return b_tec.var_tec_emissions_neg[t] == 0

            b_tec.const_tec_emissions_neg = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_neg
            )

        else:

            if emissions_based_on == "output":

                def init_tec_emissions_pos(const, t):
                    """emissions_pos = output * emissionfactor"""
                    if c["emission_factor"] >= 0:
                        return (
                            b_tec.var_output[
                                t, self.component_options.main_output_carrier
                            ]
                            * b_tec.para_tec_emissionfactor
                            == b_tec.var_tec_emissions_pos[t]
                        )
                    else:
                        return b_tec.var_tec_emissions_pos[t] == 0

                b_tec.const_tec_emissions_pos = pyo.Constraint(
                    self.set_t_global, rule=init_tec_emissions_pos
                )

                def init_tec_emissions_neg(const, t):
                    if c["emission_factor"] < 0:
                        return (
                            b_tec.var_output[
                                t, self.component_options.main_output_carrier
                            ]
                            * (-b_tec.para_tec_emissionfactor)
                            == b_tec.var_tec_emissions_neg[t]
                        )
                    else:
                        return b_tec.var_tec_emissions_neg[t] == 0

                b_tec.const_tec_emissions_neg = pyo.Constraint(
                    self.set_t_global, rule=init_tec_emissions_neg
                )

            elif emissions_based_on == "input":

                def init_tec_emissions_pos(const, t):
                    if c["emission_factor"] >= 0:
                        return (
                            b_tec.var_input[
                                t, self.component_options.main_input_carrier
                            ]
                            * b_tec.para_tec_emissionfactor
                            == b_tec.var_tec_emissions_pos[t]
                        )
                    else:
                        return b_tec.var_tec_emissions_pos[t] == 0

                b_tec.const_tec_emissions_pos = pyo.Constraint(
                    self.set_t_global, rule=init_tec_emissions_pos
                )

                def init_tec_emissions_neg(const, t):
                    if c["emission_factor"] < 0:
                        return (
                            b_tec.var_input[
                                t, self.component_options.main_input_carrier
                            ](-b_tec.para_tec_emissionfactor)
                            == b_tec.var_tec_emissions_neg[t]
                        )
                    else:
                        return b_tec.var_tec_emissions_neg[t] == 0

                b_tec.const_tec_emissions_neg = pyo.Constraint(
                    self.set_t_global, rule=init_tec_emissions_neg
                )

        return b_tec

    def _define_auxiliary_vars(self, b_tec, data: dict):
        """
        Defines auxiliary variables, that are required for the modelling of clustered data

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        c = self.processed_coeff.time_independent

        if not (self.component_options.technology_model == "RES") and not (
            self.component_options.technology_model == "CONV4"
        ):

            def init_input_bounds(bounds, t, car):
                return tuple(
                    self.bounds["input"][car][t - 1, :]
                    * c["size_max"]
                    * c["rated_power"]
                )

            b_tec.var_input_aux = pyo.Var(
                self.set_t_performance,
                b_tec.set_input_carriers,
                within=pyo.NonNegativeReals,
                bounds=init_input_bounds,
            )

            b_tec.const_link_full_resolution_input = link_full_resolution_to_clustered(
                b_tec.var_input_aux,
                b_tec.var_input,
                self.set_t_full,
                self.sequence,
                b_tec.set_input_carriers,
            )

        def init_output_bounds(bounds, t, car):
            return tuple(
                self.bounds["output"][car][t - 1, :] * c["size_max"] * c["rated_power"]
            )

        b_tec.var_output_aux = pyo.Var(
            self.set_t_performance,
            b_tec.set_output_carriers,
            within=pyo.NonNegativeReals,
            bounds=init_output_bounds,
        )

        b_tec.const_link_full_resolution_output = link_full_resolution_to_clustered(
            b_tec.var_output_aux,
            b_tec.var_output,
            self.set_t_full,
            self.sequence,
            b_tec.set_output_carriers,
        )

        return b_tec

    def _aggregate_input(self, b_tec):
        """
        Aggregates CCS and technology input

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """

        b_tec.var_input_tot = pyo.Var(
            self.set_t_global,
            b_tec.set_input_carriers_all,
            within=pyo.NonNegativeReals,
        )

        def init_aggregate_input(const, t, car):
            """input_ccs + input = input_tot"""
            input_tec = (
                b_tec.var_input[t, car] if car in b_tec.set_input_carriers else 0
            )
            if self.component_options.ccs_possible:
                input_ccs = (
                    b_tec.var_input_ccs[t, car]
                    if car in b_tec.set_input_carriers_ccs
                    else 0
                )
            else:
                input_ccs = 0
            return input_tec + input_ccs == b_tec.var_input_tot[t, car]

        b_tec.const_input_aggregation = pyo.Constraint(
            self.set_t_global, b_tec.set_input_carriers_all, rule=init_aggregate_input
        )

        return b_tec

    def _aggregate_output(self, b_tec):
        """
        Aggregates CCS and technology output

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        b_tec.var_output_tot = pyo.Var(
            self.set_t_global,
            b_tec.set_output_carriers_all,
            within=pyo.NonNegativeReals,
        )

        def init_aggregate_output(const, t, car):
            """output + output_ccs = output_tot"""
            output_tec = (
                b_tec.var_output[t, car] if car in b_tec.set_output_carriers else 0
            )
            if self.component_options.ccs_possible:
                output_ccs = (
                    b_tec.var_output_ccs[t, car]
                    if car in b_tec.set_output_carriers_ccs
                    else 0
                )
            else:
                output_ccs = 0
            return output_tec + output_ccs == b_tec.var_output_tot[t, car]

        b_tec.const_output_aggregation = pyo.Constraint(
            self.set_t_global, b_tec.set_output_carriers_all, rule=init_aggregate_output
        )

        return b_tec

    def _aggregate_cost(self, b_tec):
        """
        Aggregates CCS and technology cost

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        b_tec.var_capex_tot = pyo.Var()
        b_tec.var_opex_fixed_tot = pyo.Var()
        b_tec.var_opex_variable_tot = pyo.Var(self.set_t_global)

        def init_aggregate_capex(const):
            """capex + capex_ccs = capex_tot"""
            capex_tec = b_tec.var_capex
            if self.component_options.ccs_possible:
                capex_ccs = b_tec.var_capex_ccs
            else:
                capex_ccs = 0
            return b_tec.var_capex_tot == capex_tec + capex_ccs

        b_tec.const_capex_aggregation = pyo.Constraint(rule=init_aggregate_capex)

        def init_aggregate_opex_var(const, t):
            """var_opex_variable + var_opex_variable_ccs = var_opex_variable_tot"""
            opex_var_tec = b_tec.var_opex_variable[t]
            if self.component_options.ccs_possible:
                opex_var_ccs = b_tec.var_opex_variable_ccs[t]
            else:
                opex_var_ccs = 0
            return b_tec.var_opex_variable_tot[t] == opex_var_tec + opex_var_ccs

        b_tec.const_opex_var_aggregation = pyo.Constraint(
            self.set_t_global, rule=init_aggregate_opex_var
        )

        def init_aggregate_opex_fixed(const):
            """var_opex_fixed + var_opex_fixed_ccs = var_opex_fixed_tot"""
            opex_fixed_tec = b_tec.var_opex_fixed
            if self.component_options.ccs_possible:
                opex_fixed_ccs = b_tec.var_opex_fixed_ccs
            else:
                opex_fixed_ccs = 0
            return b_tec.var_opex_fixed_tot == opex_fixed_tec + opex_fixed_ccs

        b_tec.const_opex_fixed_aggregation = pyo.Constraint(
            rule=init_aggregate_opex_fixed
        )

        return b_tec

    def write_results_tec_design(self, h5_group, model_block):
        """
        Function to report technology design

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        h5_group.create_dataset("technology", data=[self.name])
        h5_group.create_dataset("size", data=[model_block.var_size.value])
        h5_group.create_dataset("existing", data=[self.existing])
        h5_group.create_dataset("capex_tot", data=[model_block.var_capex_tot.value])
        h5_group.create_dataset(
            "opex_variable",
            data=[
                sum(model_block.var_opex_variable[t].value for t in self.set_t_global)
            ],
        )
        h5_group.create_dataset(
            "opex_fixed_tot", data=[model_block.var_opex_fixed_tot.value]
        )
        h5_group.create_dataset(
            "emissions_pos",
            data=[
                sum(
                    model_block.var_tec_emissions_pos[t].value
                    for t in self.set_t_global
                )
            ],
        )
        h5_group.create_dataset(
            "emissions_neg",
            data=[
                sum(
                    model_block.var_tec_emissions_neg[t].value
                    for t in self.set_t_global
                )
            ],
        )
        if self.component_options.ccs_possible:
            h5_group.create_dataset("size_ccs", data=[model_block.var_size_ccs.value])
            h5_group.create_dataset("capex_tec", data=[model_block.var_capex.value])
            h5_group.create_dataset("capex_ccs", data=[model_block.var_capex_ccs.value])
            h5_group.create_dataset(
                "opex_fixed_ccs", data=[model_block.var_opex_fixed_ccs.value]
            )

        h5_group.create_dataset(
            "para_unitCAPEX", data=[model_block.para_unit_capex.value]
        )
        if hasattr(model_block, "para_fix_capex"):
            h5_group.create_dataset(
                "para_fixCAPEX", data=[model_block.para_fix_capex.value]
            )

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report technology operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """

        for car in model_block.set_input_carriers_all:
            if model_block.find_component("var_input"):
                h5_group.create_dataset(
                    f"{car}_input",
                    data=[
                        model_block.var_input_tot[t, car].value
                        for t in self.set_t_global
                    ],
                )
        for car in model_block.set_output_carriers_all:
            h5_group.create_dataset(
                f"{car}_output",
                data=[
                    model_block.var_output_tot[t, car].value for t in self.set_t_global
                ],
            )
        h5_group.create_dataset(
            "emissions_pos",
            data=[
                model_block.var_tec_emissions_pos[t].value for t in self.set_t_global
            ],
        )
        h5_group.create_dataset(
            "emissions_neg",
            data=[
                model_block.var_tec_emissions_neg[t].value for t in self.set_t_global
            ],
        )
        if model_block.find_component("var_x"):
            h5_group.create_dataset(
                "var_x",
                data=[
                    0 if x is None else x
                    for x in [
                        model_block.var_x[t].value for t in self.set_t_performance
                    ]
                ],
            )
        if model_block.find_component("var_y"):
            h5_group.create_dataset(
                "var_y",
                data=[
                    0 if x is None else x
                    for x in [
                        model_block.var_y[t].value for t in self.set_t_performance
                    ]
                ],
            )
        if model_block.find_component("var_z"):
            h5_group.create_dataset(
                "var_z",
                data=[
                    0 if x is None else x
                    for x in [
                        model_block.var_z[t].value for t in self.set_t_performance
                    ]
                ],
            )

        if model_block.find_component("set_input_carriers_ccs"):
            for car in model_block.set_input_carriers_ccs:
                h5_group.create_dataset(
                    f"{car}_var_input_ccs",
                    data=[
                        model_block.var_input_ccs[t, car].value
                        for t in self.set_t_performance
                    ],
                )
            for car in model_block.set_output_carriers_ccs:
                h5_group.create_dataset(
                    f"{car}_var_output_ccs",
                    data=[
                        model_block.var_output_ccs[t, car].value
                        for t in self.set_t_performance
                    ],
                )

    def scale_model(self, b_tec, model, config):
        """
        Scales technology model

        :param b_tec: pyomo network block
        :param model: pyomo model
        :param dict config: config dict containing scaling factors
        :return: pyomo model
        """

        f = self.scaling_factors
        f_global = config["scaling"]["scaling_factors"]

        model = determine_variable_scaling(model, b_tec, f, f_global)
        model = determine_constraint_scaling(model, b_tec, f, f_global)

        return model

    # CCS FUNCTIONS
    def _define_ccs_performance(self, b_tec, data: dict):
        """
        Defines CCS performance. The unit capex parameter is calculated from Eq. 10 of Weimann et al. 2023

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        coeff_ti = self.ccs_component.processed_coeff.time_independent

        emissions_based_on = self.component_options.emissions_based_on
        capture_rate = coeff_ti["capture_rate"]

        # Initialize the size of CCS as in _define_size (size given in mass flow of CO2 entering the CCS object)
        b_tec.para_size_min_ccs = pyo.Param(
            domain=pyo.NonNegativeReals,
            initialize=self.ccs_component.input_parameters.size_min,
            mutable=True,
        )
        b_tec.para_size_max_ccs = pyo.Param(
            domain=pyo.NonNegativeReals,
            initialize=self.ccs_component.input_parameters.size_max,
            mutable=True,
        )

        # Size CCS
        b_tec.var_size_ccs = pyo.Var(
            within=pyo.NonNegativeReals,
            bounds=(b_tec.para_size_min_ccs, b_tec.para_size_max_ccs),
        )

        # TODO: maybe make the full set of all carriers as an intersection between this set and  the others?
        # Emission Factor
        b_tec.para_tec_emissionfactor = pyo.Param(
            domain=pyo.Reals,
            initialize=self.processed_coeff.time_independent["emission_factor"],
        )
        b_tec.var_tec_emissions_pos = pyo.Var(
            self.set_t_global, within=pyo.NonNegativeReals
        )
        b_tec.var_tec_emissions_neg = pyo.Var(
            self.set_t_global, within=pyo.NonNegativeReals
        )

        def init_input_bounds(bounds, t, car):
            return tuple(
                self.ccs_component.bounds["input"][car][self.sequence[t - 1] - 1, :]
                * coeff_ti["size_max"]
            )

        b_tec.var_input_ccs = pyo.Var(
            self.set_t_global,
            b_tec.set_input_carriers_ccs,
            within=pyo.NonNegativeReals,
            bounds=init_input_bounds,
        )

        def init_output_bounds(bounds, t, car):
            return tuple(
                self.ccs_component.bounds["output"][car][self.sequence[t - 1] - 1, :]
                * coeff_ti["size_max"]
            )

        b_tec.var_output_ccs = pyo.Var(
            self.set_t_global,
            b_tec.set_output_carriers_ccs,
            within=pyo.NonNegativeReals,
            bounds=init_output_bounds,
        )

        # Input-output correlation
        def init_input_output_ccs(const, t):
            if emissions_based_on == "output":
                return (
                    b_tec.var_output_ccs[t, "CO2captured"]
                    <= capture_rate
                    * b_tec.para_tec_emissionfactor
                    * b_tec.var_output[t, self.component_options.main_output_carrier]
                )
            else:
                return (
                    b_tec.var_output_ccs[t, "CO2captured"]
                    <= capture_rate
                    * b_tec.para_tec_emissionfactor
                    * b_tec.var_input[t, self.component_options.main_input_carrier]
                )

        b_tec.const_input_output_ccs = pyo.Constraint(
            self.set_t_global, rule=init_input_output_ccs
        )

        def init_size_output_ccs(const, t):
            return b_tec.var_output_ccs[t, "CO2captured"] <= b_tec.var_size_ccs

        b_tec.const_size_output_ccs = pyo.Constraint(
            self.set_t_global, rule=init_size_output_ccs
        )

        # Electricity and heat demand CCS
        def init_input_ccs(const, t, car):
            return (
                b_tec.var_input_ccs[t, car]
                == coeff_ti["input_ratios"][car]
                * b_tec.var_output_ccs[t, "CO2captured"]
                / capture_rate
            )

        b_tec.const_input_el = pyo.Constraint(
            self.set_t_global, b_tec.set_input_carriers_ccs, rule=init_input_ccs
        )

        return b_tec

    def _define_ccs_emissions(self, b_tec):
        """
        Defines CCS performance. The unit capex parameter is calculated from Eq. 10 of Weimann et al. 2023

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        emissions_based_on = self.component_options.emissions_based_on

        # Emissions
        if emissions_based_on == "output":

            def init_tec_emissions_pos(const, t):
                return (
                    b_tec.var_output[t, self.component_options.main_output_carrier]
                    * b_tec.para_tec_emissionfactor
                    - b_tec.var_output_ccs[t, "CO2captured"]
                    == b_tec.var_tec_emissions_pos[t]
                )

            b_tec.const_tec_emissions_pos = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_pos
            )

            def init_tec_emissions_neg(const, t):
                return b_tec.var_tec_emissions_neg[t] == 0

            b_tec.const_tec_emissions_neg = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_neg
            )

        elif emissions_based_on == "input":

            def init_tec_emissions_pos(const, t):
                return (
                    b_tec.var_input[t, self.component_options.main_input_carrier]
                    * b_tec.para_tec_emissionfactor
                    - b_tec.var_output_ccs[t, "CO2captured"]
                    == b_tec.var_tec_emissions_pos[t]
                )

            b_tec.const_tec_emissions_pos = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_pos
            )

            def init_tec_emissions_neg(const, t):
                return b_tec.var_tec_emissions_neg[t] == 0

            b_tec.const_tec_emissions_neg = pyo.Constraint(
                self.set_t_global, rule=init_tec_emissions_neg
            )

        # Initialize the size of CCS as in _define_size (size given in mass flow of CO2 entering the CCS object)
        b_tec.para_size_min_ccs = pyo.Param(
            domain=pyo.NonNegativeReals,
            initialize=self.ccs_component.input_parameters.size_min,
            mutable=True,
        )
        b_tec.para_size_max_ccs = pyo.Param(
            domain=pyo.NonNegativeReals,
            initialize=self.ccs_component.input_parameters.size_max,
            mutable=True,
        )

        # Decommissioning is possible, size variable
        b_tec.var_size_ccs = pyo.Var(
            within=pyo.NonNegativeReals,
            bounds=(0, b_tec.para_size_max_ccs),
        )

        return b_tec

    def _define_ccs_costs(self, b_tec, data: dict):
        """
        Defines CCS costs

        :param b_tec: pyomo block with technology model
        :param dict data: dict containing model information
        :return: pyomo block with technology model
        """
        config = data["config"]

        # Costs
        economics = self.ccs_component.economics
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        b_tec.para_unit_capex_ccs = pyo.Param(
            domain=pyo.Reals,
            initialize=economics.capex_data["unit_capex"],
            mutable=True,
        )
        b_tec.para_unit_capex_annual_ccs = pyo.Param(
            domain=pyo.Reals,
            initialize=annualization_factor * economics.capex_data["unit_capex"],
            mutable=True,
        )

        b_tec.para_fix_capex_annual_ccs = pyo.Param(
            domain=pyo.Reals,
            initialize=annualization_factor * economics.capex_data["fix_capex"],
            mutable=True,
        )

        def calculate_max_capex_ccs():
            max_capex = (
                self.ccs_component.input_parameters.size_max
                * b_tec.para_unit_capex_annual_ccs
                + b_tec.para_fix_capex_annual_ccs
            )
            return (0, max_capex)

        b_tec.var_capex_aux_ccs = pyo.Var(bounds=calculate_max_capex_ccs())

        # capex unit commitment constraint
        self.big_m_transformation_required = 1
        s_indicators = range(0, 2)

        def init_installation(dis, ind):
            if ind == 0:  # tech not installed
                dis.const_capex_aux_ccs = pyo.Constraint(
                    expr=b_tec.var_capex_aux_ccs == 0
                )
                dis.const_not_installed_ccs = pyo.Constraint(
                    expr=b_tec.var_size_ccs == 0
                )
            else:  # tech installed
                dis.const_capex_aux_ccs = pyo.Constraint(
                    expr=b_tec.var_size_ccs * b_tec.para_unit_capex_annual_ccs
                    + b_tec.para_fix_capex_annual_ccs
                    == b_tec.var_capex_aux_ccs
                )
                dis.const_installed_ccs_sizelim_min = pyo.Constraint(
                    expr=b_tec.var_size_ccs >= b_tec.para_size_min_ccs
                )
                dis.const_installed_ccs_sizelim_max = pyo.Constraint(
                    expr=b_tec.var_size_ccs <= b_tec.para_size_max_ccs
                )

        b_tec.dis_installation_ccs = gdp.Disjunct(s_indicators, rule=init_installation)

        def bind_disjunctions(dis):
            return [b_tec.dis_installation_ccs[i] for i in s_indicators]

        b_tec.disjunction_installation_ccs = gdp.Disjunction(rule=bind_disjunctions)

        # CAPEX
        b_tec.var_capex_ccs = pyo.Var()
        b_tec.const_capex_ccs = pyo.Constraint(
            expr=b_tec.var_capex_ccs == b_tec.var_capex_aux_ccs
        )

        # FIXED OPEX
        b_tec.para_opex_fixed_ccs = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_fixed, mutable=True
        )
        b_tec.var_opex_fixed_ccs = pyo.Var()
        b_tec.const_opex_fixed_ccs = pyo.Constraint(
            expr=b_tec.var_capex_aux_ccs * b_tec.para_opex_fixed_ccs
            == b_tec.var_opex_fixed_ccs
        )

        # VARIABLE OPEX
        b_tec.para_opex_variable_ccs = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_variable, mutable=True
        )
        b_tec.var_opex_variable_ccs = pyo.Var(self.set_t_global)

        def init_opex_variable_ccs(const, t):
            return (
                b_tec.var_output_ccs[t, b_tec.set_output_carriers_ccs[1]]
                * b_tec.para_opex_variable_ccs
                == b_tec.var_opex_variable_ccs[t]
            )

        b_tec.const_opex_variable_ccs = pyo.Constraint(
            self.set_t_global, rule=init_opex_variable_ccs
        )

        return b_tec

    # DYNAMICS FUNCTIONS
    def _define_dynamics(self, b_tec, data: dict):
        """
        Selects the dynamic constraints that are required based on the technology dynamic performance parameters or the
        performance function type.

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        config = data["config"]

        log_msg = f"\t \t Adding dynamics to Technology {self.name}"
        log.info(log_msg)
        if config["optimization"]["typicaldays"]["N"]["value"] != 0:
            raise Exception("time aggregation with dynamics is not implemented")

        dynamics = self.processed_coeff.dynamics
        SU_load = dynamics["SU_load"]
        SD_load = dynamics["SD_load"]
        min_uptime = dynamics["min_uptime"]
        min_downtime = dynamics["min_downtime"]
        max_startups = dynamics["max_startups"]

        if (
            (min_uptime + min_downtime > -2)
            or (max_startups > -1)
            or (SU_load + SD_load > -2)
            or self.component_options.performance_function_type == 4
        ):
            b_tec = self._dynamics_SUSD_logic(b_tec)
        if not (self.component_options.performance_function_type == 4) and (
            SU_load + SD_load > -2
        ):
            b_tec = self._dynamics_fast_SUSD(b_tec)

        log_msg = f"\t \t Adding dynamics to Technology {self.name}"
        log.info(log_msg)

        return b_tec

    def _dynamics_SUSD_logic(self, b_tec):
        """
        Adds the startup and shutdown logic to the technology model and constrains the maximum number of startups.

        Based on Equations 4-5 in Morales-España, G., Ramírez-Elizondo, L., & Hobbs, B. F. (2017). Hidden power system
        inflexibilities imposed by traditional unit commitment formulations. Applied Energy, 191, 223–238.
        https://doi.org/10.1016/J.APENERGY.2017.01.089

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        dynamics = self.processed_coeff.dynamics

        # New variables
        b_tec.var_x = pyo.Var(
            self.set_t_performance, domain=pyo.NonNegativeReals, bounds=(0, 1)
        )
        b_tec.var_y = pyo.Var(
            self.set_t_performance, domain=pyo.NonNegativeReals, bounds=(0, 1)
        )
        b_tec.var_z = pyo.Var(
            self.set_t_performance, domain=pyo.NonNegativeReals, bounds=(0, 1)
        )

        # Check for default values
        para_names = ["SU_time", "SD_time"]
        for para in para_names:
            if dynamics[para] < 0:
                dynamics[para] = 0
                log_msg = (
                    "Using SU/SD logic constraints, parameter "
                    + str(para)
                    + "set to default value 0"
                )
                log.warning(log_msg)

        para_names = ["min_uptime", "min_downtime"]
        for para in para_names:
            if dynamics[para] < 0:
                dynamics[para] = 1
                log_msg = (
                    "Using SU/SD logic constraints, parameter "
                    + str(para)
                    + " set to default value 1"
                )
                log.warning(log_msg)

        # Collect parameters
        SU_time = dynamics["SU_time"]
        SD_time = dynamics["SD_time"]
        min_uptime = dynamics["min_uptime"]
        min_downtime = dynamics["min_downtime"] + SU_time + SD_time
        max_startups = dynamics["max_startups"]

        # Enforce startup/shutdown logic
        def init_SUSD_logic1(const, t):
            if t == 1:
                return pyo.Constraint.Skip
            else:
                return (
                    b_tec.var_x[t] - b_tec.var_x[t - 1]
                    == b_tec.var_y[t] - b_tec.var_z[t]
                )

        b_tec.const_SUSD_logic1 = pyo.Constraint(
            self.set_t_performance, rule=init_SUSD_logic1
        )

        def init_SUSD_logic2(const, t):
            if t >= min_uptime:
                return b_tec.var_y[t - min_uptime + 1] <= b_tec.var_x[t]
            else:
                return (
                    b_tec.var_y[len(self.set_t_performance) + (t - min_uptime + 1)]
                    <= b_tec.var_x[t]
                )

        b_tec.const_SUSD_logic2 = pyo.Constraint(
            self.set_t_performance, rule=init_SUSD_logic2
        )

        def init_SUSD_logic3(const, t):
            if t >= min_downtime:
                return b_tec.var_z[t - min_downtime + 1] <= 1 - b_tec.var_x[t]
            else:
                return (
                    b_tec.var_z[len(self.set_t_performance) + (t - min_downtime + 1)]
                    <= 1 - b_tec.var_x[t]
                )

        b_tec.const_SUSD_logic3 = pyo.Constraint(
            self.set_t_performance, rule=init_SUSD_logic3
        )

        # Constrain number of startups
        if not max_startups == -1:

            def init_max_startups(const):
                return (
                    sum(b_tec.var_y[t] for t in self.set_t_performance) <= max_startups
                )

            b_tec.const_max_startups = pyo.Constraint(rule=init_max_startups)

        return b_tec

    def _dynamics_fast_SUSD(self, b_tec):
        """
        Adds startup and shutdown load constraints to the model.

        Based on Equations 9-11 and 13 in Morales-España, G., Ramírez-Elizondo, L., & Hobbs, B. F. (2017). Hidden power
        system inflexibilities imposed by traditional unit commitment formulations. Applied Energy, 191, 223–238.
        https://doi.org/10.1016/J.APENERGY.2017.01.089

        :param b_tec: pyomo block with technology model
        :return: pyomo block with technology model
        """
        dynamics = self.processed_coeff.dynamics

        # Check for default values
        para_names = ["SU_load", "SD_load"]
        for para in para_names:
            if dynamics[para] < 0:
                dynamics[para] = 1
                log_msg = (
                    "Using SU/SD load constraints, parameter"
                    + str(para)
                    + "set to default value 1"
                )
                log.warning(log_msg)

        # Collect parameters
        SU_load = dynamics["SU_load"]
        SD_load = dynamics["SD_load"]
        main_car = self.component_options.main_input_carrier
        rated_power = self.input_parameters.rated_power

        # SU load limit
        s_indicators = range(0, 2)

        def init_SU_load(dis, t, ind):
            if ind == 0:  # no startup (y=0)
                dis.const_y_off = pyo.Constraint(expr=b_tec.var_y[t] == 0)

            else:  # tech in startup
                dis.const_y_on = pyo.Constraint(expr=b_tec.var_y[t] == 1)

                def init_SU_load_limit(cons):
                    if self.component_options.technology_model == "CONV3":
                        return (
                            self.input[t, main_car]
                            <= b_tec.var_size * SU_load * rated_power
                        )
                    else:
                        return (
                            sum(
                                self.input[t, car_input]
                                for car_input in b_tec.set_input_carriers
                            )
                            <= b_tec.var_size * SU_load * rated_power
                        )

                dis.const_SU_load_limit = pyo.Constraint(rule=init_SU_load_limit)

        b_tec.dis_SU_load = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_SU_load
        )

        def bind_disjunctions_SU_load(dis, t):
            return [b_tec.dis_SU_load[t, i] for i in s_indicators]

        b_tec.disjunction_SU_load = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions_SU_load
        )

        # SD load limit
        s_indicators = range(0, 2)

        def init_SD_load(dis, t, ind):
            if ind == 0:  # no shutdown (z=0)
                dis.const_z_off = pyo.Constraint(expr=b_tec.var_z[t] == 0)

            else:  # tech in shutdown
                dis.const_z_on = pyo.Constraint(expr=b_tec.var_z[t] == 1)

                def init_SD_load_limit(cons):
                    if t == 1:
                        return pyo.Constraint.Skip
                    else:
                        if self.component_options.technology_model == "CONV3":
                            return (
                                self.input[t - 1, main_car]
                                <= b_tec.var_size * SD_load * rated_power
                            )
                        else:
                            return (
                                sum(
                                    self.input[t - 1, car_input]
                                    for car_input in b_tec.set_input_carriers
                                )
                                <= b_tec.var_size * SD_load * rated_power
                            )

                dis.const_SD_load_limit = pyo.Constraint(rule=init_SD_load_limit)

        b_tec.dis_SD_load = gdp.Disjunct(
            self.set_t_performance, s_indicators, rule=init_SD_load
        )

        def bind_disjunctions_SD_load(dis, t):
            return [b_tec.dis_SD_load[t, i] for i in s_indicators]

        b_tec.disjunction_SD_load = gdp.Disjunction(
            self.set_t_performance, rule=bind_disjunctions_SD_load
        )

        return b_tec



import json
from pathlib import Path
import numpy as np
import pandas as pd
import os


def create_empty_network_matrix(nodes: list) -> pd.DataFrame:
    """
    Function creates matrix for defined nodes.

    :param list nodes: list of nodes to create matrices from
    :return: pandas data frame with nodes
    :rtype: pd.DataFrame
    """
    # construct matrix
    matrix = pd.DataFrame(
        data=np.full((len(nodes), len(nodes)), 0), index=nodes, columns=nodes
    )
    return matrix


def create_carbon_cost_data(timesteps: pd.date_range) -> pd.DataFrame:
    """
    Creates a data frame with carbon cost data

    :param pd.date_range timesteps: timesteps used as index
    :return: Data frame with columns: "price", "subsidy"
    :rtype: pd.DataFrame
    """
    carbon_cost = pd.DataFrame(index=timesteps, columns=["price", "subsidy"])
    return carbon_cost


def create_climate_data(timesteps: pd.date_range) -> pd.DataFrame:
    """
    Creates a data frame with climate data

    :param pd.date_range timesteps: timesteps used as index
    :return: Data frame with two columns "ghi", "dni", "dhi", "temp_air", "rh", "ws10", "TECHNOLOGYNAME_hydro_inflow"
    :rtype: pd.DataFrame
    """
    climate_data = pd.DataFrame(
        index=timesteps,
        columns=[
            "ghi",
            "dni",
            "dhi",
            "temp_air",
            "rh",
            "ws10",
            "TECHNOLOGYNAME_hydro_inflow",
        ],
    )
    return climate_data


def create_carrier_data(timesteps: pd.date_range) -> pd.DataFrame:
    """
    Creates a data frame with carrier data

    :param pd.date_range timesteps: timesteps used as index
    :return: Data frame with two columns "Demand", "Import limit", "Export limit", "Import price", "Export price", "Import emission factor", "Export emission factor", "Generic production",
    :rtype: pd.DataFrame
    """
    carrier_data = pd.DataFrame(
        index=timesteps,
        columns=[
            "Demand",
            "Import limit",
            "Export limit",
            "Import price",
            "Export price",
            "Import emission factor",
            "Export emission factor",
            "Generic production",
        ],
    )
    return carrier_data


def create_input_data_folder_template(base_path: Path | str):
    """
    Creates a folder structure based on the topology contained in the folder

    This function creates the input data folder structure required to organize the input data to the model.
    Note that the folder needs to already exist with a Topology.json file in it that specifies the nodes, carriers,
    timesteps, investment periods and the length of the investment period.

    You can create an examplary json template with the function `func:create_topology_template`

    :param str, Path base_path: path to folder
    """
    # Convert to Path
    if isinstance(base_path, str):
        base_path = Path(base_path)

    # Read Topology.json
    with open(base_path / "Topology.json") as json_file:
        topology = json.load(json_file)

    timesteps = pd.date_range(
        start=topology["start_date"],
        end=topology["end_date"],
        freq=topology["resolution"],
    )

    # Template jsons:
    networks = {"existing": [], "new": []}
    technologies = {"existing": {}, "new": []}
    energy_balance_options = {
        carrier: {"curtailment_possible": 0} for carrier in topology["carriers"]
    }

    # Template csvs
    carrier_data = create_carrier_data(timesteps)
    climate_data = create_climate_data(timesteps)
    carbon_cost = create_carbon_cost_data(timesteps)

    node_locations = pd.DataFrame(
        index=topology["nodes"], columns=["lon", "lat", "alt"]
    )

    # Make folder structure
    node_locations.to_csv(base_path / "NodeLocations.csv", sep=";")
    for investment_period in topology["investment_periods"]:
        (base_path / investment_period).mkdir(parents=True, exist_ok=True)

        # Networks
        with open(base_path / investment_period / "Networks.json", "w") as f:
            json.dump(networks, f, indent=4)
        (base_path / investment_period / "network_data").mkdir(
            parents=True, exist_ok=True
        )
        (base_path / investment_period / "network_topology").mkdir(
            parents=True, exist_ok=True
        )
        (base_path / investment_period / "network_topology" / "new").mkdir(
            parents=True, exist_ok=True
        )
        (base_path / investment_period / "network_topology" / "existing").mkdir(
            parents=True, exist_ok=True
        )
        empty_network_matrix = create_empty_network_matrix(topology["nodes"])
        empty_network_matrix.to_csv(
            base_path
            / investment_period
            / "network_topology"
            / "new"
            / "connection.csv",
            sep=";",
        )
        empty_network_matrix.to_csv(
            base_path / investment_period / "network_topology" / "new" / "distance.csv",
            sep=";",
        )
        empty_network_matrix.to_csv(
            base_path
            / investment_period
            / "network_topology"
            / "new"
            / "size_max_arcs.csv",
            sep=";",
        )
        empty_network_matrix.to_csv(
            base_path
            / investment_period
            / "network_topology"
            / "existing"
            / "size.csv",
            sep=";",
        )
        empty_network_matrix.to_csv(
            base_path
            / investment_period
            / "network_topology"
            / "existing"
            / "distance.csv",
            sep=";",
        )
        empty_network_matrix.to_csv(
            base_path
            / investment_period
            / "network_topology"
            / "existing"
            / "connection.csv",
            sep=";",
        )

        # Node data
        (base_path / investment_period / "node_data").mkdir(parents=True, exist_ok=True)
        for node in topology["nodes"]:
            (base_path / investment_period / "node_data" / node / "carrier_data").mkdir(
                parents=True, exist_ok=True
            )
            with open(
                base_path
                / investment_period
                / "node_data"
                / node
                / "Technologies.json",
                "w",
            ) as f:
                json.dump(technologies, f, indent=4)
            with open(
                base_path
                / investment_period
                / "node_data"
                / node
                / "carrier_data"
                / "EnergybalanceOptions.json",
                "w",
            ) as f:
                json.dump(energy_balance_options, f, indent=4)
            for carrier in topology["carriers"]:
                carrier_data.to_csv(
                    base_path
                    / investment_period
                    / "node_data"
                    / node
                    / "carrier_data"
                    / f"{carrier}.csv",
                    sep=";",
                )
            climate_data.to_csv(
                base_path / investment_period / "node_data" / node / "ClimateData.csv",
                sep=";",
            )
            carbon_cost.to_csv(
                base_path / investment_period / "node_data" / node / "CarbonCost.csv",
                sep=";",
            )
            (
                base_path / investment_period / "node_data" / node / "technology_data"
            ).mkdir(parents=True, exist_ok=True)


def initialize_topology_templates() -> dict:
    """
    Creates a topology template and returns it as a dict

    :return: topology_template
    :rtype: dict
    """
    topology_template = {
        "nodes": ["node1", "node2"],
        "carriers": ["electricity", "hydrogen"],
        "investment_periods": ["period1"],
        "start_date": "2022-01-01 00:00",
        "end_date": "2022-12-31 23:00",
        "resolution": "1h",
        "investment_period_length": 1,
    }
    return topology_template


def initialize_configuration_templates() -> dict:
    """
    Creates a configuration template and returns it as a dict

    :return: configuration_template
    :rtype: dict
    """
    configuration_template = {
        "optimization": {
            "objective": {
                "description": "String specifying the objective/type of optimization.",
                "options": [
                    "costs",
                    "emissions_pos",
                    "emissions_net",
                    "emissions_minC",
                    "costs_emissionlimit",
                    "pareto",
                ],
                "value": "costs",
            },
            "emission_limit": {
                "description": "emission limit to be enforced if objective function "
                "is costs_emissionlimit",
                "value": 0,
            },
            "monte_carlo": {
                "N": {
                    "description": "Number of Monte Carlo simulations (0 = off).",
                    "value": 0,
                },
                "type": {
                    "description": "Type of Monte Carlo simulation. For type 1 the user defines the standard "
                    "deviation and the components to vary. For type 2 the user provides a csv file "
                    "with the parameters and their min, max and reference values. ",
                    "options": ["normal_dis", "uniform_dis_from_file"],
                    "value": "normal_dis",
                },
                "sd": {
                    "description": "Value defining the range in which variables are varied in Monte Carlo simulations "
                    "(defined as the standard deviation of the original value).",
                    "value": 0.2,
                },
                "on_what": {
                    "description": "List: Defines component to vary.",
                    "options": [
                        "Technologies",
                        "Networks",
                        "Import",
                        "Export",
                    ],
                    "value": ["Technologies"],
                },
            },
            "pareto_points": {"description": "Number of Pareto points.", "value": 5},
            "timestaging": {
                "description": "Defines number of timesteps that are averaged (0 = off).",
                "value": 0,
            },
            "typicaldays": {
                "N": {
                    "description": "Determines number of typical days (0 = off).",
                    "value": 0,
                },
                "method": {
                    "description": "Determine method used for modeling technologies with typical days.",
                    "options": [1, 2],
                    "value": 2,
                },
                "technologies_with_full_res": {
                    "description": "If method 2 is chosen, list determines which "
                    "technologies are modelled at full resolution. "
                    "Should be at least all storage technologies.",
                    "options": [],
                    "value": ["RES", "STOR", "Hydro_Open"],
                },
            },
            "multiyear": {
                "description": "Enable multiyear analysis, if turned off max time horizon is 1 year.",
                "options": [0, 1],
                "value": 0,
            },
        },
        "solveroptions": {
            "solver": {
                "description": "String specifying the solver used.",
                "value": "gurobi",
            },
            "mipgap": {"description": "Value to define MIP gap.", "value": 0.001},
            "timelim": {
                "description": "Value to define time limit in hours.",
                "value": 10,
            },
            "threads": {
                "description": "Value to define number of threads (default is maximum available).",
                "value": 0,
            },
            "mipfocus": {
                "description": "Modifies high level solution strategy.",
                "options": [0, 1, 2, 3],
                "value": 0,
            },
            "nodefilestart": {
                "description": "Parameter to decide when nodes are compressed and written to disk.",
                "value": 60,
            },
            "method": {
                "description": "Defines algorithm used to solve continuous models.",
                "options": [-1, 0, 1, 2, 3, 4, 5],
                "value": -1,
            },
            "heuristics": {
                "description": "Parameter to determine amount of time spent in MIP heuristics.",
                "value": 0.05,
            },
            "presolve": {
                "description": "Controls the presolve level.",
                "options": [-1, 0, 1, 2],
                "value": -1,
            },
            "branchdir": {
                "description": "Determines which child node is explored first in the branch-and-cut.",
                "options": [-1, 0, 1],
                "value": 0,
            },
            "lpwarmstart": {
                "description": "Controls whether and how warm start information is used for LP.",
                "options": [0, 1, 2],
                "value": 0,
            },
            "intfeastol": {
                "description": "Value that determines the integer feasibility tolerance.",
                "value": 1e-05,
            },
            "feastol": {
                "description": "Value that determines feasibility for all constraints.",
                "value": 1e-06,
            },
            "numericfocus": {
                "description": "Degree of which Gurobi tries to detect and manage numeric issues.",
                "options": [0, 1, 2, 3],
                "value": 0,
            },
            "cuts": {
                "description": "Setting defining the aggressiveness of the global cut.",
                "options": [-1, 0, 1, 2, 3],
                "value": -1,
            },
        },
        "reporting": {
            "write_results": {
                "description": "Write results (h5 file and to Summary.xlsx) directly "
                "after the optimization.",
                "options": [0, 1],
                "value": 1,
            },
            "save_summary_path": {
                "description": "Path to save the summary file path to.",
                "value": "./userData/",
            },
            "save_path": {
                "description": "Option to define the save path.",
                "value": "./userData/",
            },
            "case_name": {
                "description": "Option to define a case study name that is added to the results folder name.",
                "value": -1,
            },
            "write_solution_diagnostics": {
                "description": "If 1, writes solution quality, if 2 also writes pyomo to Gurobi variable map and constraint map to file.",
                "options": [0, 1, 2],
                "value": 0,
            },
        },
        "energybalance": {
            "violation": {
                "description": "Determines the energy balance violation price (-1 is no violation allowed).",
                "value": -1,
            },
            "copperplate": {
                "description": "Determines if a copperplate approach is used.",
                "options": [0, 1],
                "value": 0,
            },
        },
        "economic": {
            "global_discountrate": {
                "description": "Determines if and which global discount rate is used. This holds for the CAPEX of all technologies and networks.",
                "value": -1,
            },
            "global_simple_capex_model": {
                "description": "Determines if the CAPEX model of technologies is set to 1 for all technologies.",
                "options": [0, 1],
                "value": 0,
            },
        },
        "performance": {
            "dynamics": {
                "description": "Determines if dynamics are used.",
                "options": [0, 1],
                "value": 0,
            }
        },
        "scaling": {
            "scaling_on": {
                "description": "Determines if the model is scaled. If 1, it uses global and component specific scaling factors.",
                "options": [0, 1],
                "value": 0,
            },
            "scaling_factors": {
                "energy_vars": {
                    "description": "Scaling factor used for all energy variables.",
                    "value": 0.001,
                },
                "cost_vars": {
                    "description": "Scaling factor used for all cost variables.",
                    "value": 0.001,
                },
                "objective": {
                    "description": "Scaling factor used for the objective function.",
                    "value": 1,
                },
            },
        },
    }

    return configuration_template


def create_optimization_templates(path: Path | str):
    """
    Creates an examplary topology and model configuration json file in the specified
    path.

    :param str/Path path: path to folder to create Topology.json
    """
    if isinstance(path, str):
        path = Path(path)

    topology_file = path / "Topology.json"
    config_file = path / "ConfigModel.json"

    # Check if the files already exist
    if topology_file.exists() or config_file.exists():
        print(
            f"Files already exist: {topology_file if topology_file.exists() else ''} {config_file if config_file.exists() else ''}"
        )
        return

    topology_template = initialize_topology_templates()
    configuration_template = initialize_configuration_templates()

    with open(topology_file, "w") as f:
        json.dump(topology_template, f, indent=4)
    with open(config_file, "w") as f:
        json.dump(configuration_template, f, indent=4)


def create_montecarlo_template_csv(base_path: Path | str):
    """
    Creates a template CSV file for the monte carlo parameters and saves it to the given path. The monte carlo can
    only be performed on economic parameters.

    The file should be filled by specifying the type ("Technologies", "Networks", "Import", "Export"), the name (
    specific technology or network name, carrier in case of import or export), and the parameter ('unit_CAPEX' or
    'fix_CAPEX' for technology, 'gamma1' 'gamma2' 'gamma3' or 'gamma4' for network and 'price' for import
    and export).

    :param str/Path path: path to folder to create Topology.json
    """
    if isinstance(base_path, str):
        base_path = Path(base_path)

    montecarlo_file = base_path / "MonteCarlo.csv"

    # Check if the file already exists
    if montecarlo_file.exists():
        print(f"File already exists: {montecarlo_file}")
        return

    data = {
        "type": [None],
        "name": [None],
        "parameter": [None],
        "min": [None],
        "ref": [None],
        "max": [None],
    }
    df = pd.DataFrame(data)

    df.to_csv(montecarlo_file, sep=";", index=False)


def show_available_networks():
    """
    Prints all available networks
    """
    tec_data_path = Path(
        os.path.join(os.path.dirname(__file__) + "/../data/network_data")
    )

    for root, dirs, files in os.walk(tec_data_path.resolve()):
        for file in files:
            print(file[:-5])


def show_available_technologies():
    """
    Prints all available technologies
    """
    tec_data_path = Path(
        os.path.join(os.path.dirname(__file__) + "/../data/technology_data")
    )

    for root, dirs, files in os.walk(tec_data_path.resolve()):
        for file in files:
            print(file[:-5])



def determine_network_energy_consumption(network_data: dict) -> bool:
    """
    Determines if there is network consumption for a network

    :param dict network_data: Network data
    :return: Get if network has energyconsumption or not
    :rtype: int
    """
    # Todo: This can be further extended to check if node is connected to network
    network_energy_consumption = False
    for netw in network_data:
        if network_data[netw].component_options.energyconsumption:
            network_energy_consumption = True

    return network_energy_consumption


def get_data_for_investment_period(
    data, investment_period: str, aggregation_model: str
) -> dict:
    """
    Gets data from DataHandle for specific investement_period. Writes it to a dict.

    :param data: data to use
    :param str investment_period: investment period
    :param str aggregation_model: aggregation type
    :return: data of respective investment period
    :rtype: dict
    """
    data_period = {}
    data_period["topology"] = data.topology
    data_period["technology_data"] = data.technology_data[investment_period]
    data_period["time_series"] = data.time_series[aggregation_model].loc[
        :, investment_period
    ]
    data_period["network_data"] = data.network_data[investment_period]
    data_period["energybalance_options"] = data.energybalance_options[investment_period]
    data_period["config"] = data.model_config
    if data.model_config["optimization"]["typicaldays"]["N"]["value"] != 0:
        data_period["k_means_specs"] = data.k_means_specs[investment_period]
        # data_period["averaged_specs"] = data.averaged_specs[investment_period]

    return data_period


def get_data_for_node(data: dict, node: str) -> dict:
    """
    Gets data from a dict for specific node. Writes it to a dict.

    :param dict data: data to use
    :param str node: node
    :return: data of respective node
    :rtype: dict
    """
    data_node = {}
    data_node["topology"] = data["topology"]
    data_node["technology_data"] = data["technology_data"][node]
    data_node["time_series"] = data["time_series"][node]
    data_node["network_data"] = data["network_data"]
    data_node["energybalance_options"] = data["energybalance_options"][node]
    data_node["config"] = data["config"]
    if data["config"]["optimization"]["typicaldays"]["N"]["value"] != 0:
        data_node["k_means_specs"] = data["k_means_specs"]
        # data_node["averaged_specs"] = data["averaged_specs"]

    return data_node



import warnings
import pvlib
from timezonefinder import TimezoneFinder
import pandas as pd
from pathlib import Path
import pyomo.environ as pyo
from scipy.interpolate import interp1d
import numpy as np

from ..technology import Technology
from ...utilities import get_attribute_from_dict


class Res(Technology):
    """
    Renewable technology with capacity factor (has no input)

    Resembles a renewable technology with no input. The capacity factors of the
    technology are determined for each individual technology type.

    So far the following renewable technologies are possible:

    - Photovoltaic (based on irradiance in climate data and PV lib)

    - Wind turbines (based on wind speed and power curves provided)

    **Constraint declarations:**

    - Output of technology. The output can be curtailed in three different ways.
      For ``curtailment == 0``, there is no curtailment possible. For ``curtailment
      == 1``, the curtailment is continuous. For ``curtailment == 2``,
      the size needs to be an integer, and the technology can only be curtailed discretely, i.e. by turning full
      modules off. For ``curtailment == 0`` (default), it holds:

    .. math::
        Output_{t, car} = CapFactor_t * Size
    """

    def __init__(self, tec_data: dict):
        """
        Constructor

        :param dict tec_data: technology data
        """
        super().__init__(tec_data)

        self.component_options.emissions_based_on = "output"

    def fit_technology_performance(self, climate_data: pd.DataFrame, location: dict):
        """
        Fits technology performance

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        """
        super(Res, self).fit_technology_performance(climate_data, location)

        if "Photovoltaic" in self.name:
            if "system_type" in self.input_parameters.performance_data:
                self._perform_fitting_pv(
                    climate_data,
                    location,
                    system_data=self.input_parameters.performance_data["system_type"],
                )
            else:
                self._perform_fitting_pv(climate_data, location)

        elif "SolarThermal" in self.name:
            self._perform_fitting_ST(climate_data)

        elif "WindTurbine" in self.name:
            if "hubheight" in self.input_parameters.performance_data:
                hubheight = self.input_parameters.performance_data["hubheight"]
            else:
                hubheight = 120
            self._perform_fitting_wt(climate_data, hubheight)

        # Options
        self.component_options.other["curtailment"] = get_attribute_from_dict(
            self.input_parameters.performance_data, "curtailment", 0
        )

    def _perform_fitting_pv(self, climate_data: pd.DataFrame, location: dict, **kwargs):
        """
        Calculates capacity factors and specific area requirements for a PV system using pvlib

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param dict location: dict containing location details
        :param PV_type: (optional) can specify a certain type of module, angle, ...
        """
        if not kwargs.__contains__("system_data"):
            system_data = dict()
            system_data["tilt"] = 18
            system_data["surface_azimuth"] = 180
            system_data["module_name"] = "SunPower_SPR_X20_327"
            system_data["inverter_eff"] = 0.96
        else:
            system_data = kwargs["system_data"]

        def define_pv_system(location: dict, system_data: dict):
            """
            defines the pv system
            :param dict location: location information (latitude, longitude, altitude,
            time zone)
            :param dict system_data: contains data on tilt, surface_azimuth,
            module_name, inverter efficiency
            :return: returns PV model chain, peak power, specific area requirements
            """
            module_database = pvlib.pvsystem.retrieve_sam("CECMod")
            module = module_database[system_data["module_name"]]

            # Define temperature losses of module
            temperature_model_parameters = (
                pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS["sapm"][
                    "open_rack_glass_glass"
                ]
            )

            # Create PV model chain
            inverter_parameters = {
                "pdc0": 5000,
                "eta_inv_nom": system_data["inverter_eff"],
            }
            system = pvlib.pvsystem.PVSystem(
                surface_tilt=system_data["tilt"],
                surface_azimuth=system_data["surface_azimuth"],
                module_parameters=module,
                inverter_parameters=inverter_parameters,
                temperature_model_parameters=temperature_model_parameters,
            )

            pv_model = pvlib.modelchain.ModelChain(
                system, location, spectral_model="no_loss", aoi_model="physical"
            )
            peakpower = module.STC
            specific_area = module.STC / module.A_c / 1000 / 1000

            return pv_model, peakpower, specific_area

        # Define parameters for convinience
        lon = location["lon"]
        lat = location["lat"]
        alt = location["alt"]

        if (
            (np.isnan(location["lon"]))
            or (np.isnan(location["lat"]))
            or (np.isnan(location["alt"]))
        ):
            raise Exception(
                "To use Photovoltaic technology you need to specify a "
                "location in the NodeLocations.csv file"
            )

        # Get location
        tf = TimezoneFinder()
        tz = tf.timezone_at(lng=lon, lat=lat)
        location = pvlib.location.Location(lat, lon, tz=tz, altitude=alt)

        # Initialize pv_system
        pv_model, peakpower, specific_area = define_pv_system(location, system_data)

        # Run system with climate data
        pv_model.run_model(climate_data)

        # Calculate cap factors
        power = pv_model.results.ac.p_mp
        capacity_factor = round(power / peakpower, 3)

        # Coefficients
        self.processed_coeff.time_dependent_full["capfactor"] = capacity_factor.values
        self.processed_coeff.time_independent["specific_area"] = specific_area

    def _perform_fitting_ST(self, climate_data: pd.DataFrame):
        """
        Calculates capacity factors and specific area requirements for a solar thermal system

        :param pd.Dataframe climate_data: dataframe containing climate data
        """
        # Todo: code this
        pass

    def _perform_fitting_wt(self, climate_data: pd.DataFrame, hubheight: float):
        """
        Calculates capacity factors for a wind turbine

        The power curves are located in ``data/technology_data/RES/WT_data``

        :param pd.Dataframe climate_data: dataframe containing climate data
        :param float hubheight: hubheight of wind turbine
        """
        # Load data for wind turbine type
        wt_path = Path(__file__).parent.parent.parent.parent
        wt_data_path = wt_path / "data/technology_data/RES/WT_data/WT_data.csv"
        wt_data = pd.read_csv(wt_data_path, delimiter=";")

        # match WT with data
        if self.name in wt_data["TurbineName"]:
            wt_data = wt_data[wt_data["TurbineName"] == self.name]
        else:
            wt_data = wt_data[wt_data["TurbineName"] == "WindTurbine_Onshore_1500"]
            warnings.warn(
                "TurbineName not in csv, standard WindTurbine_Onshore_1500 selected."
            )

        # Load wind speed and correct for height
        ws = climate_data["ws10"]

        # TODO: make power exponent choice possible
        # TODO: Make different heights possible
        alpha = 1 / 7
        # if data.node_data.windPowerExponent(node) >= 0
        #     alpha = data.node_data.windPowerExponent(node);
        # else:
        #     if data.node_data.offshore(node) == 1:
        #         alpha = 0.45;
        #     else:
        #         alpha = 1 / 7;

        if hubheight > 0:
            ws = ws * (hubheight / 10) ** alpha

        # Make power curve
        rated_power = wt_data.iloc[0]["RatedPowerkW"]
        x = np.linspace(0, 35, 71)
        y = wt_data.iloc[:, 13:84]
        y = y.to_numpy()

        f = interp1d(x, y)
        ws[ws < 0] = 0
        capacity_factor = f(ws) / rated_power

        # Coefficients
        self.processed_coeff.time_dependent_full["capfactor"] = capacity_factor[
            0
        ].round(3)
        # Rated Power
        self.input_parameters.rated_power = rated_power / 1000

    def _calculate_bounds(self):
        """
        Calculates the bounds of the variables used
        """
        super(Res, self)._calculate_bounds()

        time_steps = len(self.set_t_performance)

        # Output bounds
        lower_output_bound = np.zeros(shape=(time_steps))
        upper_output_bound = self.processed_coeff.time_dependent_used["capfactor"]
        output_bounds = np.column_stack((lower_output_bound, upper_output_bound))
        self.bounds["output"]["electricity"] = output_bounds

    def construct_tech_model(self, b_tec, data: dict, set_t_full, set_t_clustered):
        """
        Adds constraints to technology blocks for tec_type RES (renewable technology)

        :param b_tec: pyomo block with technology model
        :param dict data: data containing model configuration
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with technology model
        """
        super(Res, self).construct_tech_model(b_tec, data, set_t_full, set_t_clustered)

        # DATA OF TECHNOLOGY
        coeff_td = self.processed_coeff.time_dependent_used
        rated_power = self.input_parameters.rated_power
        curtailment = self.component_options.other["curtailment"]

        # CONSTRAINTS
        if curtailment == 0:  # no curtailment allowed (default)

            def init_input_output(const, t, c_output):
                return (
                    self.output[t, c_output]
                    == coeff_td["capfactor"][t - 1] * b_tec.var_size * rated_power
                )

            b_tec.const_input_output = pyo.Constraint(
                self.set_t_performance,
                b_tec.set_output_carriers,
                rule=init_input_output,
            )

        elif curtailment == 1:  # continuous curtailment

            def init_input_output(const, t, c_output):
                return (
                    self.output[t, c_output]
                    <= coeff_td["capfactor"][t - 1] * b_tec.var_size * rated_power
                )

            b_tec.const_input_output = pyo.Constraint(
                self.set_t_performance,
                b_tec.set_output_carriers,
                rule=init_input_output,
            )

        elif curtailment == 2:  # discrete curtailment
            b_tec.var_size_on = pyo.Var(
                self.set_t_performance,
                within=pyo.NonNegativeIntegers,
                bounds=(b_tec.para_size_min, b_tec.para_size_max),
            )

            def init_curtailed_units(const, t):
                return b_tec.var_size_on[t] <= b_tec.var_size

            b_tec.const_curtailed_units = pyo.Constraint(
                self.set_t_performance, rule=init_curtailed_units
            )

            def init_input_output(const, t, c_output):
                return (
                    self.output[t, c_output]
                    == coeff_td["capfactor"][t - 1] * b_tec.var_size_on[t] * rated_power
                )

            b_tec.const_input_output = pyo.Constraint(
                self.set_t_performance,
                b_tec.set_output_carriers,
                rule=init_input_output,
            )

        return b_tec

    def write_results_tec_design(self, h5_group, model_block):
        """
        Function to report technology design

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """

        super(Res, self).write_results_tec_design(h5_group, model_block)

        h5_group.create_dataset("rated_power", data=self.input_parameters.rated_power)

    def write_results_tec_operation(self, h5_group, model_block):
        """
        Function to report technology operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        super(Res, self).write_results_tec_operation(h5_group, model_block)

        rated_power = self.input_parameters.rated_power
        capfactor = self.processed_coeff.time_dependent_used["capfactor"]

        h5_group.create_dataset(
            "max_out",
            data=[
                capfactor[t - 1] * model_block.var_size.value * rated_power
                for t in self.set_t_performance
            ],
        )

        h5_group.create_dataset("cap_factor", data=capfactor)

        if self.component_options.other["curtailment"] == 2:
            h5_group.create_dataset(
                "units_on",
                data=[model_block.var_size_on[t].value for t in self.set_t_performance],
            )

        for car in model_block.set_output_carriers:
            h5_group.create_dataset(
                "curtailment_" + car,
                data=[
                    capfactor[t - 1] * model_block.var_size.value * rated_power
                    - model_block.var_output[t, car].value
                    for t in self.set_t_performance
                ],
            )



from ..component import ModelComponent
from ..utilities import (
    annualize,
    set_discount_rate,
    perform_disjunct_relaxation,
    determine_variable_scaling,
    determine_constraint_scaling,
)

import pandas as pd
import copy
import pyomo.environ as pyo
import pyomo.gdp as gdp

import logging

log = logging.getLogger(__name__)


class Network(ModelComponent):
    """
    Class to read and manage data for networks

    For each connection between nodes, an arc is created, with its respective cost,
    flows, losses and  consumption at nodes.

    Networks that can be used in two directions (e.g. electricity cables), are called
    bidirectional and are treated respectively with their size and costs. Other
    networks, e.g. pipelines, require two installations to be able to transport in
    two directions. As such their CAPEX is double and their size in both directions
    can be different.

    **Set declarations:**

    - ``set_netw_carrier``: Set of network carrier (i.e. only one carrier, that is
      transported in the network)
    - ``set_arcs``: Set of all arcs (from_node, to_node)
    - ``set_arcs_unique``: In case the network is bidirectional: Set of unique arcs (i.e.
      for each pair of arcs, one unique entry)
    - Furthermore for each node:

        * ``set_receives_from``: A set of nodes the node receives from
        * ``set_sends_to``: A set of nodes the node sends to

    - ``set_consumed_carriers``: In case the network has an energy consumption

    **Parameter declarations:**

    - ``para_size_min``: Min Size (for each arc)
    - ``para_size_max``: Max Size (for each arc)
    - ``para_size_initial``, var_size, var_capex: for existing networks
    - ``para_capex_gamma``: :math:`{\\gamma}_1, {\\gamma}_2, {\\gamma}_3, {\\gamma}_4` for
      CAPEX calculation (annualized from given data on up-front CAPEX, lifetime and
      discount rate)
    - ``para_opex_variable``: Variable OPEX
    - ``para_opex_fixed``: Fixed OPEX
    - ``para_decommissioning_cost``: decommissioning costs for existing networks
    - ``para_send_kflow``, ``para_send_kflowDistance``, ``para_receive_kflow``,
      ``para_receive_kflowDistance``, Parameters for energy consumption at
      receiving and sending node

    **Variable declarations:**

    - ``var_capex``: CAPEX
    - ``var_opex_variable``: Variable OPEX
    - ``var_opex_fixed``: Fixed OPEX
    - Furthermore for each node:

        * ``var_netw_emissions_pos``: positive emissins at node
        * ``var_inflow``: Inflow to node (as a sum of all inflows from other nodes)
        * ``var_outflow``: Outflow from node (as a sum of all outflows toother nodes)
        * ``var_consumption``: Consumption of other carriers (e.g. electricity
          required for compression of a gas)

    **Arc Block declaration**

    Each arc represents a connection between two nodes, and is thus indexed by (
    node_from, node_to). For each arc, the following components are defined. Each
    variable is indexed by the timestep :math:`t` (here left out for convenience).

    - Decision Variables:

        * ``var_size``: Size :math:`S`
        * ``var_flow``: Flow :math:`flow`
        * ``var_losses``: Losses :math:`loss`
        * ``var_capex``, ``var_capex_aux`` CAPEX: :math:`CAPEX`
        * ``var_opex_variable``: Variable :math:`OPEXvariable`
        * ``var_emissions``: emissions from transport/losses
        * If consumption at nodes exists for network:

          * ``var_consumption_send``: Consumption at sending node :math:`Consumption_{
            nodeFrom}`
          * ``var_consumption_receive``: Consumption at receiving node
            :math:`Consumption_{nodeTo}`

    - Constraint definitions

        * Flow losses:

          .. math::
            loss = flow * {\\mu} * D

        * Flow constraints:

          .. math::
            S * minTransport \\leq flow \\leq S

        * Consumption at sending and receiving node:

          .. math::
            Consumption_{nodeFrom} = flow * k_{1, send} + flow * distance * k_{2, send}

          .. math::
            Consumption_{nodeTo} = flow * k_{1, receive} + flow * distance * k_{2, receive}

        * CAPEX of respective arc. The CAPEX is calculated as follows (for new
          networks). Note that for existing networks, the CAPEX is zero, but the
          fixed OPEX is calculated as a fraction of a hypothetical CAPEX
          based on the existing size.

          .. math::
            CAPEX_{arc} = {\\gamma}_1 + {\\gamma}_2 * S + {\\gamma}_3 * distance + {\\gamma}_4 * S * distance

        * Variable OPEX:

          .. math::
            OPEXvariable_{arc} = CAPEX_{arc} * opex_{variable}

        * Emissions:

          .. math::
            emissions = flow * f_{emissions} + loss * f_{loss2emission}

    **Network constraint declarations**
    This part calculates variables for all respective nodes.

    - CAPEX calculation of the whole network as a sum of CAPEX of all arcs. If
      ``bidirectional_network`` is set to 1 for this network, the capex and fixed
      opex for an arc is only counted once.

    - OPEX fix, as fraction of total CAPEX

    - OPEX variable as a sum of variable OPEX for each arc

    - Total emissions as the sum of all arc emissions

    - Total inflow and outflow as a sum for each node:

      .. math::
        outflow_{node} = \\sum_{nodeTo \\in sendsto_{node}} flow_{node, nodeTo}

      .. math::
        inflow_{node} = \\sum_{nodeFrom \\in receivesFrom_{node}} flow_{nodeFrom, node} - losses_{nodeFrom, node}

    - Energy consumption of other carriers at each node.

    - If  ``bidirectional_network`` is set to 1 for this network only additional
      constraints are enforced to ensure that at each time step a flow can only be in
      one direction. For ``bidirectional_network_precise = 0``, only a cut and a
      cosntraint on the sizes of the two directions of an arc are formulated:

      .. math::
        S_{nodeFrom, nodeTo} = S_{nodeTo, nodeFrom}

      .. math::
        flow_{nodeFrom, nodeTo} + flow_{nodeTo, nodeFrom} \\lor S_{nodeTo, nodeFrom}

      For ``bidirectional_network_precise = 1`` additional disjunctions are
      formulated, thus adding binaries and complexity to the model:

      .. math::
        flow_{nodeFrom, nodeTo} = 0 \\lor flow_{nodeTo, nodeFrom} = 0

    """

    def __init__(self, netw_data: dict):
        """
        Initializes network class from network data

        :param dict netw_data: technology data
        """
        super().__init__(netw_data)

        # General information
        self.connection = []
        self.distance = []
        self.size_max_arcs = []
        self.energy_consumption = {}

        # Technology Performance
        if self.component_options.energyconsumption:
            self._calculate_energy_consumption()

        self.set_nodes = []
        self.set_t = []

        self.scaling_factors = []
        if "ScalingFactors" in netw_data:
            self.scaling_factors = netw_data["ScalingFactors"]

    def fit_network_performance(self):
        """
        Fits network performance (bounds and coefficients).
        """
        input_parameters = self.input_parameters
        time_independent = {}

        # Size
        time_independent["size_min"] = input_parameters.size_min
        if not self.existing:
            time_independent["size_max"] = input_parameters.size_max
        else:
            time_independent["size_max"] = self.size_initial
            time_independent["size_initial"] = self.size_initial

        if self.existing == 0:
            if not isinstance(self.size_max_arcs, pd.DataFrame):
                # Use max size
                time_independent["size_max_arcs"] = pd.DataFrame(
                    time_independent["size_max"],
                    index=self.distance.index,
                    columns=self.distance.columns,
                )
            else:
                time_independent["size_max_arcs"] = self.size_max_arcs
        elif self.existing == 1:
            # Use initial size
            time_independent["size_max_arcs"] = time_independent["size_initial"]

        # Emissions
        time_independent["loss2emissions"] = input_parameters.performance_data[
            "loss2emissions"
        ]
        time_independent["emissionfactor"] = input_parameters.performance_data[
            "emissionfactor"
        ]

        # Other
        time_independent["rated_power"] = input_parameters.rated_power
        time_independent["min_transport"] = input_parameters.performance_data[
            "min_transport"
        ]
        time_independent["loss"] = input_parameters.performance_data["loss"]

        # Write to self
        self.processed_coeff.time_independent = time_independent

    def construct_netw_model(
        self, b_netw, data: dict, set_nodes, set_t_full, set_t_clustered
    ):
        """
        Constructs a network as model block.

        :param b_netw: pyomo network block
        :param dict data: dict containing model information
        :param set_nodes: pyomo set containing all nodes
        :param set_t_full: pyomo set containing timesteps
        :param set_t_clustered: pyomo set containing clustered timesteps
        :return: pyomo block with network model
        """
        # LOG
        log_msg = f"\t - Constructing Network {self.name}"
        log.info(log_msg)

        # NETWORK DATA
        config = data["config"]

        # MODELING TYPICAL DAYS
        self.component_options.modelled_with_full_res = True
        self.component_options.lower_res_than_full = False
        if config["optimization"]["typicaldays"]["method"]["value"] == 1:
            self.set_t = set_t_clustered
        elif config["optimization"]["typicaldays"]["method"]["value"] == 2:
            self.set_t = set_t_full
        else:
            self.set_t = set_t_full

        self.set_nodes = set_nodes

        b_netw = self._define_possible_arcs(b_netw)

        if self.component_options.bidirectional_network:
            b_netw = self._define_unique_arcs(b_netw)

        b_netw = self._define_size(b_netw)
        b_netw = self._define_capex_parameters(b_netw, data)
        b_netw = self._define_opex_parameters(b_netw)
        b_netw = self._define_emission_vars(b_netw)
        b_netw = self._define_network_characteristics(b_netw)
        b_netw = self._define_inflow_vars(b_netw)
        b_netw = self._define_outflow_vars(b_netw)

        if self.component_options.energyconsumption:
            b_netw = self._define_energyconsumption_parameters(b_netw)

        def arc_block_init(b_arc, node_from, node_to):
            """
            Constructs each arc as a block
            """

            b_arc.big_m_transformation_required = 0
            b_arc = self._define_size_arc(b_arc, b_netw, node_from, node_to)
            b_arc = self._define_capex_variables_arc(b_arc, b_netw)
            b_arc = self._define_capex_constraints_arc(
                b_arc, b_netw, node_from, node_to
            )
            b_arc = self._define_flow(b_arc, b_netw)
            b_arc = self._define_opex_arc(b_arc, b_netw)
            b_arc = self._define_emissions_arc(b_arc, b_netw)

            if self.component_options.energyconsumption:
                b_arc = self._define_energyconsumption_arc(b_arc, b_netw)

            if b_arc.big_m_transformation_required:
                b_arc = perform_disjunct_relaxation(b_arc)

            # LOG
            log_msg = f"\t\t - Constructing Arc {node_from} - {node_to} " f"completed"
            log.info(log_msg)

        b_netw.arc_block = pyo.Block(b_netw.set_arcs, rule=arc_block_init)

        # CONSTRAINTS FOR BIDIRECTIONAL NETWORKS
        if self.component_options.bidirectional_network:
            b_netw = self._define_bidirectional_constraints(b_netw)

        b_netw = self._define_capex_total(b_netw)
        b_netw = self._define_opex_total(b_netw)
        b_netw = self._define_inflow_constraints(b_netw)
        b_netw = self._define_outflow_constraints(b_netw)
        b_netw = self._define_emission_constraints(b_netw)

        if self.component_options.energyconsumption:
            b_netw = self._define_energyconsumption_total(b_netw)

        # LOG
        log_msg = f"\t - Constructing Network {self.name} completed"
        log.info(log_msg)

        return b_netw

    def _define_possible_arcs(self, b_netw):
        """
        Define all possible arcs that have a connection

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        connection = copy.deepcopy(self.connection[:])

        def init_arcs_set(set):
            for from_node in connection:
                for to_node in connection[from_node].index:
                    if connection.at[from_node, to_node] == 1:
                        yield [from_node, to_node]

        b_netw.set_arcs = pyo.Set(initialize=init_arcs_set)

        def init_nodesIn(set, node):
            for i, j in b_netw.set_arcs:
                if j == node:
                    yield i

        b_netw.set_receives_from = pyo.Set(self.set_nodes, initialize=init_nodesIn)

        def init_nodesOut(set, node):
            for i, j in b_netw.set_arcs:
                if i == node:
                    yield j

        b_netw.set_sends_to = pyo.Set(self.set_nodes, initialize=init_nodesOut)

        return b_netw

    def _define_unique_arcs(self, b_netw):
        """
        Define arcs that are unique (one arc per direction)

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        connection = copy.deepcopy(self.connection[:])

        def init_arcs_all(set):
            for from_node in connection:
                for to_node in connection[from_node].index:
                    if connection.at[from_node, to_node] == 1:
                        connection.at[to_node, from_node] = 0
                        yield [from_node, to_node]

        b_netw.set_arcs_unique = pyo.Set(initialize=init_arcs_all)

        return b_netw

    def _define_size(self, b_netw):
        """
        Defines parameters related to network size.

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        coeff_ti = self.processed_coeff.time_independent

        b_netw.para_size_min = pyo.Param(
            domain=pyo.NonNegativeReals, initialize=coeff_ti["size_min"], mutable=True
        )

        if self.existing:
            # Parameters for initial size
            def init_size_initial(param, node_from, node_to):
                return coeff_ti["size_initial"].at[node_from, node_to]

            b_netw.para_size_initial = pyo.Param(
                b_netw.set_arcs,
                domain=pyo.NonNegativeReals,
                initialize=init_size_initial,
            )
            # Check if sizes in both direction are the same for bidirectional existing networks
            if self.component_options.bidirectional_network:
                for from_node in coeff_ti["size_initial"]:
                    for to_node in coeff_ti["size_initial"][from_node].index:
                        assert (
                            coeff_ti["size_initial"].at[from_node, to_node]
                            == coeff_ti["size_initial"].at[to_node, from_node]
                        )
        return b_netw

    def _define_capex_parameters(self, b_netw, data: dict):
        """
        Defines parameters related to technology capex.

        :param b_netw: pyomo network block
        :param dict data: dict containing model information
        :return: pyomo network block
        """

        config = data["config"]
        economics = self.economics

        # CHECK FOR GLOBAL ECONOMIC OPTIONS
        discount_rate = set_discount_rate(config, economics)
        fraction_of_year_modelled = data["topology"]["fraction_of_year_modelled"]

        # CAPEX
        annualization_factor = annualize(
            discount_rate, economics.lifetime, fraction_of_year_modelled
        )

        b_netw.para_capex_gamma1 = pyo.Param(
            domain=pyo.Reals,
            mutable=True,
            initialize=economics.capex_data["gamma1"] * annualization_factor,
        )
        b_netw.para_capex_gamma2 = pyo.Param(
            domain=pyo.Reals,
            mutable=True,
            initialize=economics.capex_data["gamma2"] * annualization_factor,
        )
        b_netw.para_capex_gamma3 = pyo.Param(
            domain=pyo.Reals,
            mutable=True,
            initialize=economics.capex_data["gamma3"] * annualization_factor,
        )
        b_netw.para_capex_gamma4 = pyo.Param(
            domain=pyo.Reals,
            mutable=True,
            initialize=economics.capex_data["gamma4"] * annualization_factor,
        )

        b_netw.var_capex = pyo.Var()

        return b_netw

    def _define_opex_parameters(self, b_netw):
        """
        Defines OPEX parameters (fixed and variable)

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        economics = self.economics

        b_netw.para_opex_variable = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_variable, mutable=True
        )
        b_netw.para_opex_fixed = pyo.Param(
            domain=pyo.Reals, initialize=economics.opex_fixed, mutable=True
        )

        b_netw.var_opex_variable = pyo.Var(self.set_t)
        b_netw.var_opex_fixed = pyo.Var()
        if self.existing:
            b_netw.para_decommissioning_cost = pyo.Param(
                domain=pyo.Reals, initialize=economics.decommission_cost, mutable=True
            )

        return b_netw

    def _define_emission_vars(self, b_netw):
        """
        Defines network emissions

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        b_netw.var_netw_emissions_pos = pyo.Var(
            self.set_t, self.set_nodes, domain=pyo.NonNegativeReals
        )

        return b_netw

    def _define_network_characteristics(self, b_netw):
        """
        Defines transported carrier, losses and minimum transport requirements

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        # Define set of transported carrier
        b_netw.set_netw_carrier = pyo.Set(
            initialize=[self.component_options.transported_carrier]
        )

        return b_netw

    def _define_inflow_vars(self, b_netw):
        """
        Defines network inflow variable (i.e. sum of inflow to one node)

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        b_netw.var_inflow = pyo.Var(
            self.set_t,
            b_netw.set_netw_carrier,
            self.set_nodes,
            domain=pyo.NonNegativeReals,
        )
        return b_netw

    def _define_outflow_vars(self, b_netw):
        """
        Defines network outflow variable (i.e. sum of outflow to one node)

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        b_netw.var_outflow = pyo.Var(
            self.set_t,
            b_netw.set_netw_carrier,
            self.set_nodes,
            domain=pyo.NonNegativeReals,
        )
        return b_netw

    def _define_energyconsumption_parameters(self, b_netw):
        """
        Constructs constraints for network energy consumption

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        # Set of consumed carriers
        b_netw.set_consumed_carriers = pyo.Set(
            initialize=list(self.energy_consumption.keys())
        )

        # Parameters
        def init_cons_send1(para, car):
            return self.energy_consumption[car]["send"]["k_flow"]

        b_netw.para_send_kflow = pyo.Param(
            b_netw.set_consumed_carriers, domain=pyo.Reals, initialize=init_cons_send1
        )

        def init_cons_send2(para, car):
            return self.energy_consumption[car]["send"]["k_flowDistance"]

        b_netw.para_send_kflowDistance = pyo.Param(
            b_netw.set_consumed_carriers, domain=pyo.Reals, initialize=init_cons_send2
        )

        def init_cons_receive1(para, car):
            return self.energy_consumption[car]["receive"]["k_flow"]

        b_netw.para_receive_kflow = pyo.Param(
            b_netw.set_consumed_carriers,
            domain=pyo.Reals,
            initialize=init_cons_receive1,
        )

        def init_cons_receive2(para, car):
            return self.energy_consumption[car]["receive"]["k_flowDistance"]

        b_netw.para_receive_kflowDistance = pyo.Param(
            b_netw.set_consumed_carriers,
            domain=pyo.Reals,
            initialize=init_cons_receive2,
        )

        # Consumption at each node
        b_netw.var_consumption = pyo.Var(
            self.set_t,
            b_netw.set_consumed_carriers,
            self.set_nodes,
            domain=pyo.NonNegativeReals,
        )

        return b_netw

    def _calculate_energy_consumption(self):
        """
        Fits the performance parameters for a network, i.e. the consumption at each node.
        """
        # Get energy consumption at nodes form file
        energycons = self.input_parameters.performance_data["energyconsumption"]

        for car in energycons:
            self.energy_consumption[car] = {}
            if energycons[car]["cons_model"] == 1:
                self.energy_consumption[car]["send"] = {}
                self.energy_consumption[car]["send"] = energycons[car]
                self.energy_consumption[car]["send"].pop("cons_model")
                self.energy_consumption[car]["receive"] = {}
                self.energy_consumption[car]["receive"]["k_flow"] = 0
                self.energy_consumption[car]["receive"]["k_flowDistance"] = 0
            elif energycons[car]["cons_model"] == 2:
                temp = energycons[car]
                self.energy_consumption[car]["send"] = {}
                self.energy_consumption[car]["send"]["k_flow"] = round(
                    temp["c"]
                    * temp["T"]
                    / temp["eta"]
                    / temp["LHV"]
                    * ((temp["p"] / 30) ** ((temp["gam"] - 1) / temp["gam"]) - 1),
                    4,
                )
                self.energy_consumption[car]["send"]["k_flowDistance"] = 0
                self.energy_consumption[car]["receive"] = {}
                self.energy_consumption[car]["receive"]["k_flow"] = 0
                self.energy_consumption[car]["receive"]["k_flowDistance"] = 0

    def _define_size_arc(self, b_arc, b_netw, node_from: str, node_to: str):
        """
        Defines the size of an arc

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :param str node_from: node from which arc comes
        :param str node_to: node to which arc goes
        :return: pyomo arc block
        """
        coeff_ti = self.processed_coeff.time_independent

        if self.component_options.size_is_int:
            size_domain = pyo.NonNegativeIntegers
        else:
            size_domain = pyo.NonNegativeReals

        b_arc.para_size_max = pyo.Param(
            domain=size_domain,
            initialize=coeff_ti["size_max_arcs"].at[node_from, node_to],
        )

        b_arc.distance = self.distance.at[node_from, node_to]

        if self.existing:
            # Existing network
            if not self.component_options.decommission:
                # Decommissioning not possible
                b_arc.var_size = pyo.Param(
                    domain=size_domain,
                    initialize=b_netw.para_size_initial[node_from, node_to],
                )
            else:
                # Decommissioning possible
                b_arc.var_size = pyo.Var(
                    domain=size_domain,
                    bounds=(
                        b_netw.para_size_min,
                        b_netw.para_size_initial[node_from, node_to],
                    ),
                )
        else:
            # New network
            b_arc.var_size = pyo.Var(
                domain=size_domain, bounds=(b_netw.para_size_min, b_arc.para_size_max)
            )

        return b_arc

    def _define_capex_variables_arc(self, b_arc, b_netw):
        """
        Defines the capex variables of an arc

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :return: pyomo arc block
        """
        rated_capacity = self.input_parameters.rated_power

        def calculate_max_capex():
            max_capex = (
                b_netw.para_capex_gamma1
                + b_netw.para_capex_gamma2 * b_arc.para_size_max
                + b_netw.para_capex_gamma3 * b_arc.distance
                + b_netw.para_capex_gamma4 * b_arc.para_size_max * b_arc.distance
            )
            return (0, max_capex)

        # CAPEX auxilliary (used to calculate theoretical CAPEX)
        # For new technologies, this is equal to actual CAPEX
        # For existing technologies it is used to calculate fixed OPEX
        b_arc.var_capex_aux = pyo.Var(bounds=calculate_max_capex())

        if self.existing and not self.component_options.decommission:
            b_arc.var_capex = pyo.Param(domain=pyo.NonNegativeReals, initialize=0)
        else:
            b_arc.var_capex = pyo.Var(bounds=calculate_max_capex())

        return b_arc

    def _define_capex_constraints_arc(self, b_arc, b_netw, node_from, node_to):
        """
        Defines the capex of an arc and corresponding constraints

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :param str node_from: node from which arc comes
        :param str node_to: node to which arc goes
        :return: pyomo arc block
        """
        rated_capacity = self.input_parameters.rated_power

        def init_capex(const):
            return (
                b_arc.var_capex_aux
                == b_netw.para_capex_gamma1
                + b_netw.para_capex_gamma2 * b_arc.var_size
                + b_netw.para_capex_gamma3 * b_arc.distance
                + b_netw.para_capex_gamma4 * b_arc.var_size * b_arc.distance
            )

        # CAPEX aux:
        if self.existing and not self.component_options.decommission:
            b_arc.const_capex_aux = pyo.Constraint(rule=init_capex)
        elif (b_netw.para_capex_gamma1.value == 0) and (
            b_netw.para_capex_gamma3.value == 0
        ):
            b_arc.const_capex_aux = pyo.Constraint(rule=init_capex)
        else:
            b_arc.big_m_transformation_required = 1
            s_indicators = range(0, 2)

            def init_installation(dis, ind):
                if ind == 0:  # network not installed
                    dis.const_capex_aux = pyo.Constraint(expr=b_arc.var_capex_aux == 0)
                    dis.const_not_installed = pyo.Constraint(expr=b_arc.var_size == 0)
                else:  # network installed
                    dis.const_capex_aux = pyo.Constraint(rule=init_capex)

            b_arc.dis_installation = gdp.Disjunct(s_indicators, rule=init_installation)

            def bind_disjunctions(dis):
                return [b_arc.dis_installation[i] for i in s_indicators]

            b_arc.disjunction_installation = gdp.Disjunction(rule=bind_disjunctions)

        # CAPEX and CAPEX aux
        if self.existing and self.component_options.decommission:
            b_arc.const_capex = pyo.Constraint(
                expr=b_arc.var_capex
                == (b_netw.para_size_initial[node_from, node_to] - b_arc.var_size)
                * b_netw.para_decommissioning_cost
            )
        elif not self.existing:
            b_arc.const_capex = pyo.Constraint(
                expr=b_arc.var_capex == b_arc.var_capex_aux
            )

        return b_arc

    def _define_flow(self, b_arc, b_netw):
        """
        Defines the flow through one arc and respective losses

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :return: pyomo arc block
        """
        rated_capacity = self.input_parameters.rated_power
        coeff_ti = self.processed_coeff.time_independent

        b_arc.var_flow = pyo.Var(
            self.set_t,
            domain=pyo.NonNegativeReals,
            bounds=(
                b_netw.para_size_min * rated_capacity,
                b_arc.para_size_max * rated_capacity,
            ),
        )
        b_arc.var_losses = pyo.Var(
            self.set_t,
            domain=pyo.NonNegativeReals,
            bounds=(
                b_netw.para_size_min * rated_capacity,
                b_arc.para_size_max * rated_capacity,
            ),
        )

        # Losses
        def init_flowlosses(const, t):
            return (
                b_arc.var_losses[t]
                == b_arc.var_flow[t] * coeff_ti["loss"] * b_arc.distance
            )

        b_arc.const_flowlosses = pyo.Constraint(self.set_t, rule=init_flowlosses)

        # Flow-size-constraint
        def init_size_const_high(const, t):
            return b_arc.var_flow[t] <= b_arc.var_size * rated_capacity

        b_arc.const_flow_size_high = pyo.Constraint(
            self.set_t, rule=init_size_const_high
        )

        def init_size_const_low(const, t):
            return (
                b_arc.var_size * rated_capacity * coeff_ti["min_transport"]
                <= b_arc.var_flow[t]
            )

        b_arc.const_flow_size_low = pyo.Constraint(self.set_t, rule=init_size_const_low)
        return b_arc

    def _define_energyconsumption_arc(self, b_arc, b_netw):
        """
        Defines the energyconsumption for an arc

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :return: pyomo arc block
        """
        rated_capacity = self.input_parameters.rated_power

        b_arc.var_consumption_send = pyo.Var(
            self.set_t,
            b_netw.set_consumed_carriers,
            domain=pyo.NonNegativeReals,
            bounds=(b_netw.para_size_min, b_arc.para_size_max * rated_capacity),
        )
        b_arc.var_consumption_receive = pyo.Var(
            self.set_t,
            b_netw.set_consumed_carriers,
            domain=pyo.NonNegativeReals,
            bounds=(b_netw.para_size_min, b_arc.para_size_max * rated_capacity),
        )

        # Sending node
        def init_consumption_send(const, t, car):
            return (
                b_arc.var_consumption_send[t, car]
                == b_arc.var_flow[t] * b_netw.para_send_kflow[car]
                + b_arc.var_flow[t]
                * b_netw.para_send_kflowDistance[car]
                * b_arc.distance
            )

        b_arc.const_consumption_send = pyo.Constraint(
            self.set_t, b_netw.set_consumed_carriers, rule=init_consumption_send
        )

        # Receiving node
        def init_consumption_receive(const, t, car):
            return (
                b_arc.var_consumption_receive[t, car]
                == b_arc.var_flow[t] * b_netw.para_receive_kflow[car]
                + b_arc.var_flow[t]
                * b_netw.para_receive_kflowDistance[car]
                * b_arc.distance
            )

        b_arc.const_consumption_receive = pyo.Constraint(
            self.set_t, b_netw.set_consumed_carriers, rule=init_consumption_receive
        )

        return b_arc

    def _define_opex_arc(self, b_arc, b_netw):
        """
        Defines OPEX per Arc

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :return: pyomo arc block
        """
        b_arc.var_opex_variable = pyo.Var(self.set_t)

        def init_opex_variable(const, t):
            return (
                b_arc.var_opex_variable[t]
                == b_arc.var_flow[t] * b_netw.para_opex_variable
            )

        b_arc.const_opex_variable = pyo.Constraint(self.set_t, rule=init_opex_variable)
        return b_arc

    def _define_emissions_arc(self, b_arc, b_netw):
        """
        defines emission per arc

        :param b_arc: pyomo arc block
        :param b_netw: pyomo network block
        :return: pyomo arc block
        """
        coeff_ti = self.processed_coeff.time_independent

        b_arc.var_emissions = pyo.Var(self.set_t)

        def init_arc_emissions(const, t):
            return (
                b_arc.var_emissions[t]
                == b_arc.var_flow[t] * coeff_ti["emissionfactor"]
                + b_arc.var_losses[t] * coeff_ti["loss2emissions"]
            )

        b_arc.const_arc_emissions = pyo.Constraint(self.set_t, rule=init_arc_emissions)

        return b_arc

    def _define_bidirectional_constraints(self, b_netw):
        """
        Defines constraints necessary, in case one arc can transport in two directions.

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        rated_capacity = self.input_parameters.rated_power

        # Size in both direction is the same
        if self.component_options.decommission or not self.existing:

            def init_size_bidirectional(const, node_from, node_to):
                return (
                    b_netw.arc_block[node_from, node_to].var_size
                    == b_netw.arc_block[node_to, node_from].var_size
                )

            b_netw.const_size_bidirectional = pyo.Constraint(
                b_netw.set_arcs_unique, rule=init_size_bidirectional
            )

        s_indicators = range(0, 2)

        # Cut according to Germans work
        def init_cut_bidirectional(const, t, node_from, node_to):
            return (
                b_netw.arc_block[node_from, node_to].var_flow[t]
                + b_netw.arc_block[node_to, node_from].var_flow[t]
                <= b_netw.arc_block[node_from, node_to].var_size * rated_capacity
            )

        b_netw.const_cut_bidirectional = pyo.Constraint(
            self.set_t, b_netw.set_arcs_unique, rule=init_cut_bidirectional
        )

        # Disjunction
        if self.component_options.bidirectional_network_precise:
            self.big_m_transformation_required = 1

            def init_bidirectional(dis, t, node_from, node_to, ind):
                if ind == 0:

                    def init_bidirectional1(const):
                        return b_netw.arc_block[node_from, node_to].var_flow[t] == 0

                    dis.const_flow_zero = pyo.Constraint(rule=init_bidirectional1)

                else:

                    def init_bidirectional2(const):
                        return b_netw.arc_block[node_to, node_from].var_flow[t] == 0

                    dis.const_flow_zero = pyo.Constraint(rule=init_bidirectional2)

            b_netw.dis_one_direction_only = gdp.Disjunct(
                self.set_t,
                b_netw.set_arcs_unique,
                s_indicators,
                rule=init_bidirectional,
            )

            # Bind disjuncts
            def bind_disjunctions(dis, t, node_from, node_to):
                return [
                    b_netw.dis_one_direction_only[t, node_from, node_to, i]
                    for i in s_indicators
                ]

            b_netw.disjunction_one_direction_only = gdp.Disjunction(
                self.set_t, b_netw.set_arcs_unique, rule=bind_disjunctions
            )

        return b_netw

    def _define_capex_total(self, b_netw):
        """
        Defines total CAPEX of network

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        if self.component_options.bidirectional_network:
            arc_set = b_netw.set_arcs_unique
        else:
            arc_set = b_netw.set_arcs

        def init_capex(const):
            return (
                sum(b_netw.arc_block[arc].var_capex for arc in arc_set)
                == b_netw.var_capex
            )

        b_netw.const_capex = pyo.Constraint(rule=init_capex)

        return b_netw

    def _define_opex_total(self, b_netw):
        """
        Defines total OPEX of network

        :param b_netw: pyomo network block
        :return: pyomo network block
        """
        if self.component_options.bidirectional_network:
            arc_set = b_netw.set_arcs_unique
        else:
            arc_set = b_netw.set_arcs

        def init_opex_fixed(const):
            return (
                b_netw.para_opex_fixed
                * sum(b_netw.arc_block[arc].var_capex_aux for arc in arc_set)
                == b_netw.var_opex_fixed
            )

        b_netw.const_opex_fixed = pyo.Constraint(rule=init_opex_fixed)

        def init_opex_variable(const, t):
            return (
                sum(
                    b_netw.arc_block[arc].var_opex_variable[t]
                    for arc in b_netw.set_arcs
                )
                == b_netw.var_opex_variable[t]
            )

        b_netw.const_opex_var = pyo.Constraint(self.set_t, rule=init_opex_variable)
        return b_netw

    def _define_inflow_constraints(self, b_netw):
        """
        Connects the arc flows to inflow at each node

        :param b_netw: pyomo network block
        :return: pyomo network block
        """

        def init_inflow(const, t, car, node):
            return b_netw.var_inflow[t, car, node] == sum(
                b_netw.arc_block[from_node, node].var_flow[t]
                - b_netw.arc_block[from_node, node].var_losses[t]
                for from_node in b_netw.set_receives_from[node]
            )

        b_netw.const_inflow = pyo.Constraint(
            self.set_t, b_netw.set_netw_carrier, self.set_nodes, rule=init_inflow
        )
        return b_netw

    def _define_outflow_constraints(self, b_netw):
        """
        Connects the arc flows to outflow at each node

        :param b_netw: pyomo network block
        :return: pyomo network block
        """

        def init_outflow(const, t, car, node):
            return b_netw.var_outflow[t, car, node] == sum(
                b_netw.arc_block[node, to_node].var_flow[t]
                for to_node in b_netw.set_sends_to[node]
            )

        b_netw.const_outflow = pyo.Constraint(
            self.set_t, b_netw.set_netw_carrier, self.set_nodes, rule=init_outflow
        )
        return b_netw

    def _define_emission_constraints(self, b_netw):
        """
        Defines emissions from network

        :param b_netw: pyomo network block
        :return: pyomo network block
        """

        def init_netw_emissions(const, t, node):
            return b_netw.var_netw_emissions_pos[t, node] == sum(
                b_netw.arc_block[from_node, node].var_emissions[t]
                for from_node in b_netw.set_receives_from[node]
            )

        b_netw.const_netw_emissions = pyo.Constraint(
            self.set_t, self.set_nodes, rule=init_netw_emissions
        )
        return b_netw

    def _define_energyconsumption_total(self, b_netw):
        """
        Defines network consumption at each node

        :param b_netw: pyomo network block
        :return: pyomo network block
        """

        def init_network_consumption(const, t, car, node):
            return b_netw.var_consumption[t, car, node] == sum(
                b_netw.arc_block[node, to_node].var_consumption_send[t, car]
                for to_node in b_netw.set_sends_to[node]
            ) + sum(
                b_netw.arc_block[from_node, node].var_consumption_receive[t, car]
                for from_node in b_netw.set_receives_from[node]
            )

        b_netw.const_netw_consumption = pyo.Constraint(
            self.set_t,
            b_netw.set_consumed_carriers,
            self.set_nodes,
            rule=init_network_consumption,
        )

        return b_netw

    def write_results_netw_design(self, h5_group, model_block):
        """
        Function to report network design

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        coeff_ti = self.processed_coeff.time_independent

        for arc_name in model_block.set_arcs:
            arc = model_block.arc_block[arc_name]
            str = "".join(arc_name)
            arc_group = h5_group.create_group(str)

            arc_group.create_dataset(
                "para_capex_gamma1", data=model_block.para_capex_gamma1.value
            )
            arc_group.create_dataset(
                "para_capex_gamma2", data=model_block.para_capex_gamma2.value
            )
            arc_group.create_dataset(
                "para_capex_gamma3", data=model_block.para_capex_gamma3.value
            )
            arc_group.create_dataset(
                "para_capex_gamma4", data=model_block.para_capex_gamma4.value
            )
            arc_group.create_dataset("network", data=self.name)
            arc_group.create_dataset("fromNode", data=arc_name[0])
            arc_group.create_dataset("toNode", data=arc_name[1])
            arc_group.create_dataset("size", data=arc.var_size.value)
            arc_group.create_dataset("capex", data=arc.var_capex.value)
            arc_group.create_dataset(
                "opex_fixed",
                data=[model_block.para_opex_fixed.value * arc.var_capex_aux.value],
            )
            arc_group.create_dataset(
                "opex_variable",
                data=sum(arc.var_opex_variable[t].value for t in self.set_t),
            )
            arc_group.create_dataset(
                "total_flow", data=sum(arc.var_flow[t].value for t in self.set_t)
            )
            total_emissions = (
                sum(arc.var_flow[t].value for t in self.set_t)
                * coeff_ti["emissionfactor"]
                + sum(arc.var_losses[t].value for t in self.set_t)
                * coeff_ti["loss2emissions"]
            )
            arc_group.create_dataset("total_emissions", data=total_emissions)

    def write_results_netw_operation(self, h5_group, model_block):
        """
        Function to report network operation

        :param model_block: pyomo network block
        :param h5_group: h5 group to write to
        """
        for arc_name in model_block.set_arcs:
            arc = model_block.arc_block[arc_name]
            str = "".join(arc_name)
            arc_group = h5_group.create_group(str)

            arc_group.create_dataset(
                "flow", data=[arc.var_flow[t].value for t in self.set_t]
            )
            arc_group.create_dataset(
                "losses", data=[arc.var_losses[t].value for t in self.set_t]
            )

            if arc.find_component("var_consumption_send"):
                for car in model_block.set_consumed_carriers:

                    arc_group.create_dataset(
                        "consumption_send" + car,
                        data=[
                            arc.var_consumption_send[t, car].value for t in self.set_t
                        ],
                    )
                    arc_group.create_dataset(
                        "consumption_receive" + car,
                        data=[
                            arc.var_consumption_receive[t, car].value
                            for t in self.set_t
                        ],
                    )

    def scale_model(self, b_netw, model, config: dict):
        """
        Scales network model

        :param b_netw: pyomo network block
        :param model: pyomo model
        :param dict config: config dict containing scaling factors
        :return: pyomo model
        """

        f = self.scaling_factors
        f_global = config["scaling"]["scaling_factors"]

        model = determine_variable_scaling(model, b_netw, f, f_global)
        model = determine_constraint_scaling(model, b_netw, f, f_global)

        for arc in b_netw.arc_block:
            b_arc = b_netw.arc_block[arc]

            model = determine_variable_scaling(model, b_arc, f, f_global)
            model = determine_constraint_scaling(model, b_arc, f, f_global)

        return model
